@@ -24,7 +24,12 @@ typedef ssize_t(*http_send_callback)(const uint8_t *buffer,
 				     const size_t buffer_len,
 				     struct session *session);
 
-typedef queue_t(int32_t) queue_int32_t;
+struct http_stream {
+	int32_t id;
+	kr_http_header_array_t *headers;
+};
+
+typedef queue_t(struct http_stream) queue_http_stream;
 
 typedef enum {
 	HTTP_METHOD_NONE = 0,
@@ -36,11 +41,12 @@ struct http_ctx {
 	struct nghttp2_session *h2;
 	http_send_callback send_cb;
 	struct session *session;
-	queue_int32_t streams;  /* IDs of streams present in the buffer. */
+	queue_http_stream streams;  /* Streams present in the wire buffer. */
 	int32_t incomplete_stream;
 	ssize_t submitted;
 	http_method_t current_method;
 	char *uri_path;
+	kr_http_header_array_t *headers;
 	uint8_t *buf;  /* Part of the wire_buf that belongs to current HTTP/2 stream. */
 	ssize_t buf_pos;
 	ssize_t buf_size;
@@ -52,4 +58,5 @@ ssize_t http_process_input_data(struct session *session, const uint8_t *buf, ssi
 int http_write(uv_write_t *req, uv_handle_t *handle, knot_pkt_t* pkt, int32_t stream_id,
 	       uv_write_cb on_write);
 void http_free(struct http_ctx *ctx);
+void http_free_headers(kr_http_header_array_t *headers);
 #endif
