@@ -36,16 +36,11 @@ ffi.cdef[[
  * Data structures
  */
 
-/* stdlib */
-typedef long time_t;
-struct timeval {
-	time_t tv_sec;
-	time_t tv_usec;
-};
 struct sockaddr {
     uint16_t sa_family;
     uint8_t _stub[]; /* Do not touch */
 };
+
 struct knot_error {
 	int code;
 };
@@ -59,40 +54,6 @@ int inet_pton(int af, const char *src, void *dst);
 int gettimeofday(struct timeval *tv, struct timezone *tz);
 ]]
 
-
--- TMP: compatibility with both libknot 2.8 and 2.9
-local knot_rdataset_t_cdef
-local sover_pos = string.find(libknot_SONAME, '%d')
-if not sover_pos then
-	error('unexpected libknot soname: ' .. libknot_SONAME)
-end
-local sover = string.sub(libknot_SONAME, sover_pos , sover_pos)
-if sover == '9' then
-	knot_rdataset_t_cdef = [[
-		typedef struct {
-			uint16_t count;
-			knot_rdata_t *rdata;
-		} knot_rdataset_t;
-	]]
-elseif sover == '1' then -- it's 10 really, but this is simpler
-	knot_rdataset_t_cdef = [[
-		typedef struct {
-			uint16_t count;
-			uint32_t size;
-			knot_rdata_t *rdata;
-		} knot_rdataset_t;
-	]]
-else
-	error('unexpected libknot version: ' .. sover)
-end
-ffi.cdef([[
-		typedef struct {
-			uint16_t len;
-			uint8_t data[];
-		} knot_rdata_t;
-	]] .. knot_rdataset_t_cdef)
-
-
 require('kres-gen')
 
 -- Error code representation
@@ -750,14 +711,14 @@ ffi.metatype( kr_query_t, {
 local function void_xpcall_log_tb(func, req, msg)
 	local ok, err = xpcall(func, debug.traceback, req, msg)
 	if not ok then
-		log('error: callback %s req %s msg %s stack traceback:\n%s', func, req, msg, err)
+		log_error(ffi.C.LOG_GRP_SYSTEM, 'callback %s req %s msg %s stack traceback:\n%s', func, req, msg, err)
 	end
 end
 
 local function void_xpcall_finish_tb(func, req)
 	local ok, err = xpcall(func, debug.traceback, req)
 	if not ok then
-		log('error: callback %s req %s stack traceback:\n%s', func, req, err)
+		log_error(ffi.C.LOG_GRP_SYSTEM, 'callback %s req %s stack traceback:\n%s', func, req, err)
 	end
 end
 
