@@ -17,15 +17,16 @@
 #include "lib/dnssec.h"
 #include "lib/dnssec/ta.h"
 #include "lib/resolve.h"
+#include "lib/log.h"
 
 #include <arpa/inet.h>
-#include <assert.h>
 #include <getopt.h>
 #include <libgen.h>
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/resource.h>
+#include <sys/wait.h>
 #include <unistd.h>
 
 #if ENABLE_CAP_NG
@@ -44,8 +45,20 @@ struct args the_args_value;  /** Static allocation for the_args singleton. */
 
 static void signal_handler(uv_signal_t *handle, int signum)
 {
-	uv_stop(uv_default_loop());
-	uv_signal_stop(handle);
+	switch (signum) {
+	case SIGINT:  /* Fallthrough. */
+	case SIGTERM:
+		uv_stop(uv_default_loop());
+		uv_signal_stop(handle);
+		break;
+	case SIGCHLD:
+		/* Wait for all dead processes. */
+		while (waitpid(-1, NULL, WNOHANG) > 0);
+		break;
+	default:
+		kr_log_error(SYSTEM, "unhandled signal: %d\n", signum);
+		break;
+	}
 }
 
 /** SIGBUS -> attempt to remove the overflowing cache file and abort. */
@@ -117,11 +130,7 @@ static void help(int argc, char *argv[])
 	       " -c, --config=[path]    Config file path (relative to [rundir]) (default: config).\n"
 	       " -n, --noninteractive   Don't start the read-eval-print loop for stdin+stdout.\n"
 	       " -q, --quiet            No command prompt in interactive mode.\n"
-	       " -v, --verbose          Run in verbose mode."
-#ifdef NOVERBOSELOG
-	           " (Recompile without -DNOVERBOSELOG to activate.)"
-#endif
-	           "\n"
+	       " -v, --verbose          Increase logging to debug level.\n"
 	       " -V, --version        Print version of the server.\n"
 	       " -h, --help           Print help and usage.\n"
 	       "Options:\n"
@@ -140,8 +149,8 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, bool leader, struc
 	case UV_NAMED_PIPE:	/* echo 'quit()' | kresd ... */
 		break;
 	default:
-		kr_log_error(
-			"[system] error: standard input is not a terminal or pipe; "
+		kr_log_error(SYSTEM,
+			"error: standard input is not a terminal or pipe; "
 			"use '-n' if you want non-interactive mode.  "
 			"Commands can be simply added to your configuration file or sent over the control socket.\n"
 			);
@@ -153,7 +162,7 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, bool leader, struc
 	uv_pipe_init(loop, pipe, 0);
 	if (args->interactive) {
 		if (!args->quiet)
-			printf("[system] interactive mode\n> ");
+			printf("Interactive mode:\n" "> ");
 		pipe->data = io_tty_alloc_data();
 		uv_pipe_open(pipe, 0);
 		uv_read_start((uv_stream_t*)pipe, io_tty_alloc, io_tty_process_input);
@@ -197,14 +206,6 @@ static void args_deinit(struct args *args)
 	array_clear(args->config);
 }
 
-static long strtol_10(const char *s)
-{
-	if (!s) abort();
-	/* ^^ This shouldn't ever happen.  When getopt_long() returns an option
-	 * character that has a mandatory parameter, optarg can't be NULL. */
-	return strtol(s, NULL, 10);
-}
-
 /** Process arguments into struct args.
  * @return >=0 if main() should be exited immediately.
  */
@@ -229,24 +230,27 @@ static int parse_args(int argc, char **argv, struct args *args)
 		switch (c)
 		{
 		case 'a':
+			kr_require(optarg);
 			array_push(args->addrs, optarg);
 			break;
 		case 't':
+			kr_require(optarg);
 			array_push(args->addrs_tls, optarg);
 			break;
 		case 'c':
-			assert(optarg != NULL);
+			kr_require(optarg);
 			array_push(args->config, optarg);
 			break;
 		case 'f':
-			args->forks = strtol_10(optarg);
+			kr_require(optarg);
+			args->forks = strtol(optarg, NULL, 10);
 			if (args->forks == 1) {
-				kr_log_deprecate("use --noninteractive instead of --forks=1\n");
+				kr_log_deprecate(SYSTEM, "use --noninteractive instead of --forks=1\n");
 			} else {
-				kr_log_deprecate("support for running multiple --forks will be removed\n");
+				kr_log_deprecate(SYSTEM, "support for running multiple --forks will be removed\n");
 			}
 			if (args->forks <= 0) {
-				kr_log_error("[system] error '-f' requires a positive"
+				kr_log_error(SYSTEM, "error '-f' requires a positive"
 						" number, not '%s'\n", optarg);
 				return EXIT_FAILURE;
 			}
@@ -255,16 +259,13 @@ static int parse_args(int argc, char **argv, struct args *args)
 			args->interactive = false;
 			break;
 		case 'v':
-			kr_verbose_set(true);
-#ifdef NOVERBOSELOG
-			kr_log_info("--verbose flag has no effect due to compilation with -DNOVERBOSELOG.\n");
-#endif
+			kr_log_level_set(LOG_DEBUG);
 			break;
 		case 'q':
 			args->quiet = true;
 			break;
 		case 'V':
-			kr_log_info("%s, version %s\n", "Knot Resolver", PACKAGE_VERSION);
+			kr_log_info(SYSTEM, "%s, version %s\n", "Knot Resolver", PACKAGE_VERSION);
 			return EXIT_SUCCESS;
 		case 'h':
 		case '?':
@@ -274,7 +275,7 @@ static int parse_args(int argc, char **argv, struct args *args)
 			help(argc, argv);
 			return EXIT_FAILURE;
 		case 'S':
-			(void)0;
+			kr_require(optarg);
 			flagged_fd_t ffd = { 0 };
 			char *endptr;
 			ffd.fd = strtol(optarg, &endptr, 10);
@@ -289,7 +290,7 @@ static int parse_args(int argc, char **argv, struct args *args)
 				/* Some other kind; no checks here. */
 				ffd.flags.kind = strdup(endptr + 1);
 			} else {
-				kr_log_error("[system] incorrect value passed to '-S/--fd': %s\n",
+				kr_log_error(SYSTEM, "incorrect value passed to '-S/--fd': %s\n",
 						optarg);
 				return EXIT_FAILURE;
 			}
@@ -348,7 +349,7 @@ static int bind_sockets(addr_array_t *addrs, bool tls, flagged_fd_array_t *fds)
 		}
 		free(sa);
 		if (ret != 0) {
-			kr_log_error("[system] bind to '%s'%s: %s\n",
+			kr_log_error(NETWORK, "bind to '%s'%s: %s\n",
 				addrs->at[i], tls ? " (TLS)" : "", kr_strerror(ret));
 			has_error = true;
 		}
@@ -367,7 +368,7 @@ static int start_listening(struct network *net, flagged_fd_array_t *fds) {
 			/* TODO: try logging address@port.  It's not too important,
 			 * because typical problems happen during binding already.
 			 * (invalid address, permission denied) */
-			kr_log_error("[system] listen on fd=%d: %s\n",
+			kr_log_error(NETWORK, "listen on fd=%d: %s\n",
 					ffd->fd, kr_strerror(ret));
 			/* Continue printing all of these before exiting. */
 		} else {
@@ -383,7 +384,7 @@ static void drop_capabilities(void)
 #if ENABLE_CAP_NG
 	/* Drop all capabilities when running under non-root user. */
 	if (geteuid() == 0) {
-		kr_log_verbose("[system] running as root, no capabilities dropped\n");
+		kr_log_debug(SYSTEM, "running as root, no capabilities dropped\n");
 		return;
 	}
 	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
@@ -391,14 +392,14 @@ static void drop_capabilities(void)
 
 		/* Apply. */
 		if (capng_apply(CAPNG_SELECT_BOTH) < 0) {
-			kr_log_error("[system] failed to set process capabilities: %s\n",
+			kr_log_error(SYSTEM, "failed to set process capabilities: %s\n",
 			          strerror(errno));
 		} else {
-			kr_log_verbose("[system] all capabilities dropped\n");
+			kr_log_debug(SYSTEM, "all capabilities dropped\n");
 		}
 	} else {
 		/* If user() was called, the capabilities were already dropped along with SETPCAP. */
-		kr_log_verbose("[system] process not allowed to set capabilities, skipping\n");
+		kr_log_debug(SYSTEM, "process not allowed to set capabilities, skipping\n");
 	}
 #endif /* ENABLE_CAP_NG */
 }
@@ -406,12 +407,12 @@ static void drop_capabilities(void)
 int main(int argc, char **argv)
 {
 	if (setvbuf(stdout, NULL, _IONBF, 0) || setvbuf(stderr, NULL, _IONBF, 0)) {
-		kr_log_error("[system] failed to to set output buffering (ignored): %s\n",
+		kr_log_error(SYSTEM, "failed to to set output buffering (ignored): %s\n",
 				strerror(errno));
 		fflush(stderr);
 	}
 	if (strcmp("linux", OPERATING_SYSTEM) != 0)
-		kr_log_info("[warn] Knot Resolver is tested on Linux, other platforms might exhibit bugs.\n"
+		kr_log_warning(SYSTEM, "Knot Resolver is tested on Linux, other platforms might exhibit bugs.\n"
 				"Please report issues to https://gitlab.nic.cz/knot/knot-resolver/issues/\n"
 				"Thank you for your time and interest!\n");
 
@@ -430,7 +431,7 @@ int main(int argc, char **argv)
 		/* FIXME: access isn't a good way if we start as root and drop privileges later */
 		if (access(the_args->rundir, W_OK) != 0
 		    || chdir(the_args->rundir) != 0) {
-			kr_log_error("[system] rundir '%s': %s\n",
+			kr_log_error(SYSTEM, "rundir '%s': %s\n",
 					the_args->rundir, strerror(errno));
 			return EXIT_FAILURE;
 		}
@@ -448,7 +449,7 @@ int main(int argc, char **argv)
 		} else if (access(config, R_OK) != 0) {
 			char cwd[PATH_MAX];
 			get_workdir(cwd, sizeof(cwd));
-			kr_log_error("[system] config '%s' (workdir '%s'): %s\n",
+			kr_log_error(SYSTEM, "config '%s' (workdir '%s'): %s\n",
 				config, cwd, strerror(errno));
 			return EXIT_FAILURE;
 		}
@@ -463,16 +464,16 @@ int main(int argc, char **argv)
 	struct rlimit rlim;
 	ret = getrlimit(RLIMIT_NOFILE, &rlim);
 	if (ret == 0 && rlim.rlim_cur != rlim.rlim_max) {
-		kr_log_verbose("[system] increasing file-descriptor limit: %ld -> %ld\n",
+		kr_log_debug(SYSTEM, "increasing file-descriptor limit: %ld -> %ld\n",
 				(long)rlim.rlim_cur, (long)rlim.rlim_max);
 		rlim.rlim_cur = rlim.rlim_max;
 		ret = setrlimit(RLIMIT_NOFILE, &rlim);
 	}
 	if (ret) {
-		kr_log_error("[system] failed to get or set file-descriptor limit: %s\n",
+		kr_log_error(SYSTEM, "failed to get or set file-descriptor limit: %s\n",
 				strerror(errno));
 	} else if (rlim.rlim_cur < 512*1024) {
-		kr_log_info("[system] warning: hard limit for number of file-descriptors is only %ld but recommended value is 524288\n",
+		kr_log_warning(SYSTEM, "warning: hard limit for number of file-descriptors is only %ld but recommended value is 524288\n",
 				(long)rlim.rlim_cur);
 	}
 
@@ -490,23 +491,25 @@ int main(int argc, char **argv)
 	static struct engine engine;
 	ret = engine_init(&engine, &pool);
 	if (ret != 0) {
-		kr_log_error("[system] failed to initialize engine: %s\n", kr_strerror(ret));
+		kr_log_error(SYSTEM, "failed to initialize engine: %s\n", kr_strerror(ret));
 		return EXIT_FAILURE;
 	}
 	/* Initialize the worker */
 	ret = worker_init(&engine, the_args->forks);
 	if (ret != 0) {
-		kr_log_error("[system] failed to initialize worker: %s\n", kr_strerror(ret));
+		kr_log_error(SYSTEM, "failed to initialize worker: %s\n", kr_strerror(ret));
 		return EXIT_FAILURE;
 	}
 
 	uv_loop_t *loop = uv_default_loop();
 	/* Catch some signals. */
-	uv_signal_t sigint, sigterm;
+	uv_signal_t sigint, sigterm, sigchld;
 	if (true) ret = uv_signal_init(loop, &sigint);
 	if (!ret) ret = uv_signal_init(loop, &sigterm);
+	if (!ret) ret = uv_signal_init(loop, &sigchld);
 	if (!ret) ret = uv_signal_start(&sigint, signal_handler, SIGINT);
 	if (!ret) ret = uv_signal_start(&sigterm, signal_handler, SIGTERM);
+	if (!ret) ret = uv_signal_start(&sigchld, signal_handler, SIGCHLD);
 	/* Block SIGPIPE; see https://github.com/libuv/libuv/issues/45 */
 	if (!ret && signal(SIGPIPE, SIG_IGN) == SIG_ERR) ret = errno;
 	if (!ret) {
@@ -521,7 +524,7 @@ int main(int argc, char **argv)
 		}
 	}
 	if (ret) {
-		kr_log_error("[system] failed to set up signal handlers: %s\n",
+		kr_log_error(SYSTEM, "failed to set up signal handlers: %s\n",
 				strerror(abs(errno)));
 		ret = EXIT_FAILURE;
 		goto cleanup;
@@ -530,7 +533,7 @@ int main(int argc, char **argv)
 	 * (of the usual type) may skew results, e.g. epoll_pwait() taking lots of time. */
 	ret = uv_loop_configure(loop, UV_LOOP_BLOCK_SIGNAL, SIGPROF);
 	if (ret) {
-		kr_log_info("[system] failed to block SIGPROF in event loop, ignoring: %s\n",
+		kr_log_info(SYSTEM, "failed to block SIGPROF in event loop, ignoring: %s\n",
 				uv_strerror(ret));
 	}
 
@@ -542,7 +545,7 @@ int main(int argc, char **argv)
 
 	ret = udp_queue_init_global(loop);
 	if (ret) {
-		kr_log_error("[system] failed to initialize UDP queue: %s\n",
+		kr_log_error(SYSTEM, "failed to initialize UDP queue: %s\n",
 				kr_strerror(ret));
 		ret = EXIT_FAILURE;
 		goto cleanup;
