@@ -14,7 +14,6 @@
 #include <gnutls/x509.h>
 #include <uv.h>
 
-#include <assert.h>
 #include <errno.h>
 #include <stdlib.h>
 
@@ -28,9 +27,15 @@
 #define EPHEMERAL_CERT_EXPIRATION_SECONDS_RENEW_BEFORE (60*60*24*7)
 #define GNUTLS_PIN_MIN_VERSION  0x030400
 
+#define VERBOSE_MSG(cl_side, ...)\
+	if (cl_side) \
+		kr_log_debug(TLSCLIENT, __VA_ARGS__); \
+	else \
+		kr_log_debug(TLS, __VA_ARGS__);
+
 /** @internal Debugging facility. */
 #ifdef DEBUG
-#define DEBUG_MSG(...) kr_log_verbose("[tls] " __VA_ARGS__)
+#define DEBUG_MSG(...) kr_log_debug(TLS, __VA_ARGS__)
 #else
 #define DEBUG_MSG(...)
 #endif
@@ -41,9 +46,6 @@ struct async_write_ctx {
 	char buf[];
 };
 
-static char const server_logstring[] = "tls";
-static char const client_logstring[] = "tls_client";
-
 static int client_verify_certificate(gnutls_session_t tls_session);
 
 /**
@@ -61,7 +63,7 @@ static int kres_gnutls_set_priority(gnutls_session_t session) {
 	const char *errpos = NULL;
 	int err = gnutls_priority_set_direct(session, priorities, &errpos);
 	if (err != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] setting priority '%s' failed at character %zd (...'%s') with %s (%d)\n",
+		kr_log_error(TLS, "setting priority '%s' failed at character %zd (...'%s') with %s (%d)\n",
 			     priorities, errpos - priorities, errpos, gnutls_strerror_name(err), err);
 	}
 	return err;
@@ -70,7 +72,10 @@ static int kres_gnutls_set_priority(gnutls_session_t session) {
 static ssize_t kres_gnutls_pull(gnutls_transport_ptr_t h, void *buf, size_t len)
 {
 	struct tls_common_ctx *t = (struct tls_common_ctx *)h;
-	assert(t != NULL);
+	if (kr_fails_assert(t)) {
+		errno = EFAULT;
+		return -1;
+	}
 
 	ssize_t	avail = t->nread - t->consumed;
 	DEBUG_MSG("[%s] pull wanted: %zu available: %zu\n",
@@ -88,11 +93,14 @@ static ssize_t kres_gnutls_pull(gnutls_transport_ptr_t h, void *buf, size_t len)
 
 static void on_write_complete(uv_write_t *req, int status)
 {
-	assert(req->data != NULL);
+	if (kr_fails_assert(req->data))
+		return;
 	struct async_write_ctx *async_ctx = (struct async_write_ctx *)req->data;
 	struct tls_common_ctx *t = async_ctx->t;
-	assert(t->write_queue_size);
-	t->write_queue_size -= 1;
+	if (t->write_queue_size)
+		t->write_queue_size -= 1;
+	else
+		kr_assert(false);
 	free(req->data);
 }
 
@@ -104,8 +112,7 @@ static bool stream_queue_is_empty(struct tls_common_ctx *t)
 static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * iov, int iovcnt)
 {
 	struct tls_common_ctx *t = (struct tls_common_ctx *)h;
-
-	if (t == NULL) {
+	if (kr_fails_assert(t)) {
 		errno = EFAULT;
 		return -1;
 	}
@@ -114,9 +121,15 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
 		return 0;
 	}
 
-	assert(t->session);
+	if (kr_fails_assert(t->session)) {
+		errno = EFAULT;
+		return -1;
+	}
 	uv_stream_t *handle = (uv_stream_t *)session_get_handle(t->session);
-	assert(handle && handle->type == UV_TCP);
+	if (kr_fails_assert(handle && handle->type == UV_TCP)) {
+		errno = EFAULT;
+		return -1;
+	}
 
 	/*
 	 * This is a little bit complicated. There are two different writes:
@@ -155,8 +168,8 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
 		if (ret < 0 && ret != UV_EAGAIN) {
 			/* uv_try_write() has returned error code other then UV_EAGAIN.
 			 * Return. */
-			kr_log_verbose("[%s] uv_try_write error: %s\n",
-				       t->client_side ? "tls_client" : "tls", uv_strerror(ret));
+			VERBOSE_MSG(t->client_side, "uv_try_write error: %s\n",
+					uv_strerror(ret));
 			ret = -1;
 			errno = EIO;
 			return ret;
@@ -214,8 +227,8 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
 			t->write_queue_size += 1;
 		} else {
 			free(p);
-			kr_log_verbose("[%s] uv_write error: %s\n",
-				       t->client_side ? "tls_client" : "tls", uv_strerror(ret));
+			VERBOSE_MSG(t->client_side, "uv_write error: %s\n",
+					uv_strerror(ret));
 			errno = EIO;
 			ret = -1;
 		}
@@ -236,15 +249,14 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
   */
 static int tls_handshake(struct tls_common_ctx *ctx, tls_handshake_cb handshake_cb) {
 	struct session *session = ctx->session;
-	const char *logstring = ctx->client_side ? client_logstring : server_logstring;
 
 	int err = gnutls_handshake(ctx->tls_session);
 	if (err == GNUTLS_E_SUCCESS) {
 		/* Handshake finished, return success */
 		ctx->handshake_state = TLS_HS_DONE;
 		struct sockaddr *peer = session_get_peer(session);
-		kr_log_verbose("[%s] TLS handshake with %s has completed\n",
-			       logstring,  kr_straddr(peer));
+		VERBOSE_MSG(ctx->client_side, "TLS handshake with %s has completed\n",
+				kr_straddr(peer));
 		if (handshake_cb) {
 			if (handshake_cb(session, 0) != kr_ok()) {
 				return kr_error(EIO);
@@ -254,9 +266,8 @@ static int tls_handshake(struct tls_common_ctx *ctx, tls_handshake_cb handshake_
 		return kr_error(EAGAIN);
 	} else if (gnutls_error_is_fatal(err)) {
 		/* Fatal errors, return error as it's not recoverable */
-		kr_log_verbose("[%s] gnutls_handshake failed: %s (%d)\n",
-			     logstring,
-		             gnutls_strerror_name(err), err);
+		VERBOSE_MSG(ctx->client_side, "gnutls_handshake failed: %s (%d)\n",
+				gnutls_strerror_name(err), err);
 		/* Notify the peer about handshake failure via an alert. */
 		gnutls_alert_send_appropriate(ctx->tls_session, err);
 		if (handshake_cb) {
@@ -268,8 +279,8 @@ static int tls_handshake(struct tls_common_ctx *ctx, tls_handshake_cb handshake_
 		const char *alert_name = gnutls_alert_get_name(gnutls_alert_get(ctx->tls_session));
 		if (alert_name != NULL) {
 			struct sockaddr *peer = session_get_peer(session);
-			kr_log_verbose("[%s] TLS alert from %s received: %s\n",
-				       logstring, kr_straddr(peer), alert_name);
+			VERBOSE_MSG(ctx->client_side, "TLS alert from %s received: %s\n",
+					kr_straddr(peer), alert_name);
 		}
 	}
 	return kr_ok();
@@ -278,17 +289,17 @@ static int tls_handshake(struct tls_common_ctx *ctx, tls_handshake_cb handshake_
 
 struct tls_ctx *tls_new(struct worker_ctx *worker)
 {
-	assert(worker != NULL);
-	assert(worker->engine != NULL);
+	if (kr_fails_assert(worker && worker->engine))
+		return NULL;
 
 	struct network *net = &worker->engine->net;
 	if (!net->tls_credentials) {
 		net->tls_credentials = tls_get_ephemeral_credentials(worker->engine);
 		if (!net->tls_credentials) {
-			kr_log_error("[tls] X.509 credentials are missing, and ephemeral credentials failed; no TLS\n");
+			kr_log_error(TLS, "X.509 credentials are missing, and ephemeral credentials failed; no TLS\n");
 			return NULL;
 		}
-		kr_log_info("[tls] Using ephemeral TLS credentials\n");
+		kr_log_info(TLS, "Using ephemeral TLS credentials\n");
 		tls_credentials_log_pins(net->tls_credentials);
 	}
 
@@ -301,15 +312,15 @@ struct tls_ctx *tls_new(struct worker_ctx *worker)
 				if (newcreds) {
 					tls_credentials_release(net->tls_credentials);
 					net->tls_credentials = newcreds;
-					kr_log_info("[tls] Renewed expiring ephemeral X.509 cert\n");
+					kr_log_info(TLS, "Renewed expiring ephemeral X.509 cert\n");
 				} else {
-					kr_log_error("[tls] Failed to renew expiring ephemeral X.509 cert, using existing one\n");
+					kr_log_error(TLS, "Failed to renew expiring ephemeral X.509 cert, using existing one\n");
 				}
 			}
 		} else {
 			/* non-ephemeral cert: warn once when certificate expires */
 			if (now >= net->tls_credentials->valid_until) {
-				kr_log_error("[tls] X.509 certificate has expired!\n");
+				kr_log_error(TLS, "X.509 certificate has expired!\n");
 				net->tls_credentials->valid_until = GNUTLS_X509_NO_WELL_DEFINED_EXPIRATION;
 			}
 		}
@@ -317,13 +328,13 @@ struct tls_ctx *tls_new(struct worker_ctx *worker)
 
 	struct tls_ctx *tls = calloc(1, sizeof(struct tls_ctx));
 	if (tls == NULL) {
-		kr_log_error("[tls] failed to allocate TLS context\n");
+		kr_log_error(TLS, "failed to allocate TLS context\n");
 		return NULL;
 	}
 
 	int err = gnutls_init(&tls->c.tls_session, GNUTLS_SERVER | GNUTLS_NONBLOCK);
 	if (err != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] gnutls_init(): %s (%d)\n", gnutls_strerror_name(err), err);
+		kr_log_error(TLS, "gnutls_init(): %s (%d)\n", gnutls_strerror_name(err), err);
 		tls_free(tls);
 		return NULL;
 	}
@@ -331,7 +342,7 @@ struct tls_ctx *tls_new(struct worker_ctx *worker)
 	err = gnutls_credentials_set(tls->c.tls_session, GNUTLS_CRD_CERTIFICATE,
 				     tls->credentials->credentials);
 	if (err != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] gnutls_credentials_set(): %s (%d)\n", gnutls_strerror_name(err), err);
+		kr_log_error(TLS, "gnutls_credentials_set(): %s (%d)\n", gnutls_strerror_name(err), err);
 		tls_free(tls);
 		return NULL;
 	}
@@ -357,16 +368,12 @@ struct tls_ctx *tls_new(struct worker_ctx *worker)
 
 void tls_close(struct tls_common_ctx *ctx)
 {
-	if (ctx == NULL || ctx->tls_session == NULL) {
+	if (ctx == NULL || ctx->tls_session == NULL || kr_fails_assert(ctx->session))
 		return;
-	}
-
-	assert(ctx->session);
 
 	if (ctx->handshake_state == TLS_HS_DONE) {
 		const struct sockaddr *peer = session_get_peer(ctx->session);
-		kr_log_verbose("[%s] closing tls connection to `%s`\n",
-			       ctx->client_side ? "tls_client" : "tls",
+		VERBOSE_MSG(ctx->client_side, "closing tls connection to `%s`\n",
 			       kr_straddr(peer));
 		ctx->handshake_state = TLS_HS_CLOSING;
 		gnutls_bye(ctx->tls_session, GNUTLS_SHUT_RDWR);
@@ -398,19 +405,18 @@ int tls_write(uv_write_t *req, uv_handle_t *handle, knot_pkt_t *pkt, uv_write_cb
 	struct session *s = handle->data;
 	struct tls_common_ctx *tls_ctx = session_tls_get_common_ctx(s);
 
-	assert (tls_ctx);
-	assert (session_flags(s)->outgoing == tls_ctx->client_side);
+	if (kr_fails_assert(tls_ctx && session_flags(s)->outgoing == tls_ctx->client_side))
+		return kr_error(EINVAL);
 
 	const uint16_t pkt_size = htons(pkt->size);
-	const char *logstring = tls_ctx->client_side ? client_logstring : server_logstring;
 	gnutls_session_t tls_session = tls_ctx->tls_session;
 
 	gnutls_record_cork(tls_session);
 	ssize_t count = 0;
 	if ((count = gnutls_record_send(tls_session, &pkt_size, sizeof(pkt_size)) < 0) ||
 	    (count = gnutls_record_send(tls_session, pkt->wire, pkt->size) < 0)) {
-		kr_log_verbose("[%s] gnutls_record_send failed: %s (%zd)\n",
-			       logstring, gnutls_strerror_name(count), count);
+		VERBOSE_MSG(tls_ctx->client_side, "gnutls_record_send failed: %s (%zd)\n",
+				gnutls_strerror_name(count), count);
 		return kr_error(EIO);
 	}
 
@@ -421,15 +427,14 @@ int tls_write(uv_write_t *req, uv_handle_t *handle, knot_pkt_t *pkt, uv_write_cb
 		if (!gnutls_error_is_fatal(ret)) {
 			return kr_error(EAGAIN);
 		} else {
-			kr_log_verbose("[%s] gnutls_record_uncork failed: %s (%d)\n",
-				       logstring, gnutls_strerror_name(ret), ret);
+			VERBOSE_MSG(tls_ctx->client_side, "gnutls_record_uncork failed: %s (%d)\n",
+					gnutls_strerror_name(ret), ret);
 			return kr_error(EIO);
 		}
 	}
 
 	if (ret != submitted) {
-		kr_log_error("[%s] gnutls_record_uncork didn't send all data (%d of %zd)\n",
-		             logstring, ret, submitted);
+		kr_log_error(TLS, "gnutls_record_uncork didn't send all data (%d of %zd)\n", ret, submitted);
 		return kr_error(EIO);
 	}
 
@@ -447,15 +452,11 @@ ssize_t tls_process_input_data(struct session *s, const uint8_t *buf, ssize_t nr
 		return kr_error(ENOSYS);
 	}
 
-	assert(tls_p->session == s);
+	if (kr_fails_assert(tls_p->session == s))
+		return kr_error(EINVAL);
 	const bool ok = tls_p->recv_buf == buf && nread <= sizeof(tls_p->recv_buf);
-	if (!ok) {
-		assert(false);
-		/* don't risk overflowing the buffer if we have a mistake somewhere */
+	if (kr_fails_assert(ok)) /* don't risk overflowing the buffer if we have a mistake somewhere */
 		return kr_error(EINVAL);
-	}
-
-	const char *logstring = tls_p->client_side ? client_logstring : server_logstring;
 
 	tls_p->buf = buf;
 	tls_p->nread = nread >= 0 ? nread : 0;
@@ -489,8 +490,8 @@ ssize_t tls_process_input_data(struct session *s, const uint8_t *buf, ssize_t nr
 		} else if (count == GNUTLS_E_REHANDSHAKE) {
 			/* See https://www.gnutls.org/manual/html_node/Re_002dauthentication.html */
 			struct sockaddr *peer = session_get_peer(s);
-			kr_log_verbose("[%s] TLS rehandshake with %s has started\n",
-				       logstring,  kr_straddr(peer));
+			VERBOSE_MSG(tls_p->client_side, "TLS rehandshake with %s has started\n",
+					kr_straddr(peer));
 			tls_set_hs_state(tls_p, TLS_HS_IN_PROGRESS);
 			int err = kr_ok();
 			while (tls_p->handshake_state <= TLS_HS_IN_PROGRESS) {
@@ -508,13 +509,13 @@ ssize_t tls_process_input_data(struct session *s, const uint8_t *buf, ssize_t nr
 			/* There are can be data available, check it. */
 			continue;
 		} else if (count < 0) {
-			kr_log_verbose("[%s] gnutls_record_recv failed: %s (%zd)\n",
-				     logstring, gnutls_strerror_name(count), count);
+			VERBOSE_MSG(tls_p->client_side, "gnutls_record_recv failed: %s (%zd)\n",
+					gnutls_strerror_name(count), count);
 			return kr_error(EIO);
 		} else if (count == 0) {
 			break;
 		}
-		DEBUG_MSG("[%s] received %zd data\n", logstring, count);
+		DEBUG_MSG("[%s] received %zd data\n", tls_p->client_side ? "tls_client" : "tls", count);
 		wire_buf += count;
 		wire_buf_size -= count;
 		submitted += count;
@@ -555,8 +556,7 @@ ssize_t tls_process_input_data(struct session *s, const uint8_t *buf, ssize_t nr
  * \return error code */
 static int get_oob_key_pin(gnutls_x509_crt_t crt, char *outchar, ssize_t outchar_len, bool raw)
 {
-	if (raw && outchar_len < TLS_SHA256_RAW_LEN) {
-		assert(false);
+	if (kr_fails_assert(!raw || outchar_len >= TLS_SHA256_RAW_LEN)) {
 		return kr_error(ENOSPC);
 		/* With !raw we have check inside kr_base64_encode. */
 	}
@@ -582,8 +582,7 @@ static int get_oob_key_pin(gnutls_x509_crt_t crt, char *outchar, ssize_t outchar
 	if (err >= 0 && err < outchar_len) {
 		err = GNUTLS_E_SUCCESS;
 		outchar[err] = '\0'; /* kr_base64_encode() doesn't do it */
-	} else if (err >= 0) {
-		assert(false);
+	} else if (kr_fails_assert(err < 0)) {
 		err = kr_error(ENOSPC); /* base64 fits but '\0' doesn't */
 		outchar[outchar_len - 1] = '\0';
 	}
@@ -602,7 +601,7 @@ void tls_credentials_log_pins(struct tls_credentials *tls_credentials)
 							index, &certs, &cert_count);
 		if (err != GNUTLS_E_SUCCESS) {
 			if (err != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
-				kr_log_error("[tls] could not get X.509 certificates (%d) %s\n",
+				kr_log_error(TLS, "could not get X.509 certificates (%d) %s\n",
 						err, gnutls_strerror_name(err));
 			}
 			return;
@@ -612,10 +611,10 @@ void tls_credentials_log_pins(struct tls_credentials *tls_credentials)
 			char pin[TLS_SHA256_BASE64_BUFLEN] = { 0 };
 			err = get_oob_key_pin(certs[i], pin, sizeof(pin), false);
 			if (err != GNUTLS_E_SUCCESS) {
-				kr_log_error("[tls] could not calculate RFC 7858 OOB key-pin from cert %d (%d) %s\n",
+				kr_log_error(TLS, "could not calculate RFC 7858 OOB key-pin from cert %d (%d) %s\n",
 						i, err, gnutls_strerror_name(err));
 			} else {
-				kr_log_info("[tls] RFC 7858 OOB key-pin (%d): pin-sha256=\"%s\"\n",
+				kr_log_info(TLS, "RFC 7858 OOB key-pin (%d): pin-sha256=\"%s\"\n",
 						i, pin);
 			}
 			gnutls_x509_crt_deinit(certs[i]);
@@ -626,7 +625,7 @@ void tls_credentials_log_pins(struct tls_credentials *tls_credentials)
 #else
 void tls_credentials_log_pins(struct tls_credentials *tls_credentials)
 {
-	kr_log_verbose("[tls] could not calculate RFC 7858 OOB key-pin; GnuTLS 3.4.0+ required\n");
+	kr_log_debug(TLS, "could not calculate RFC 7858 OOB key-pin; GnuTLS 3.4.0+ required\n");
 }
 #endif
 
@@ -650,17 +649,17 @@ static time_t _get_end_entity_expiration(gnutls_certificate_credentials_t creds)
 	time_t ret = GNUTLS_X509_NO_WELL_DEFINED_EXPIRATION;
 
 	if ((err = gnutls_certificate_get_crt_raw(creds, 0, 0, &data)) != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] failed to get cert to check expiration: (%d) %s\n",
+		kr_log_error(TLS, "failed to get cert to check expiration: (%d) %s\n",
 			     err, gnutls_strerror_name(err));
 		goto done;
 	}
 	if ((err = gnutls_x509_crt_init(&cert)) != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] failed to initialize cert: (%d) %s\n",
+		kr_log_error(TLS, "failed to initialize cert: (%d) %s\n",
 			     err, gnutls_strerror_name(err));
 		goto done;
 	}
 	if ((err = gnutls_x509_crt_import(cert, &data, GNUTLS_X509_FMT_DER)) != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] failed to construct cert while checking expiration: (%d) %s\n",
+		kr_log_error(TLS, "failed to construct cert while checking expiration: (%d) %s\n",
 			     err, gnutls_strerror_name(err));
 		goto done;
 	}
@@ -686,14 +685,14 @@ int tls_certificate_set(struct network *net, const char *tls_cert, const char *t
 
 	int err = 0;
 	if ((err = gnutls_certificate_allocate_credentials(&tls_credentials->credentials)) != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls] gnutls_certificate_allocate_credentials() failed: (%d) %s\n",
+		kr_log_error(TLS, "gnutls_certificate_allocate_credentials() failed: (%d) %s\n",
 			     err, gnutls_strerror_name(err));
 		tls_credentials_free(tls_credentials);
 		return kr_error(ENOMEM);
 	}
 	if ((err = gnutls_certificate_set_x509_system_trust(tls_credentials->credentials)) < 0) {
 		if (err != GNUTLS_E_UNIMPLEMENTED_FEATURE) {
-			kr_log_error("[tls] warning: gnutls_certificate_set_x509_system_trust() failed: (%d) %s\n",
+			kr_log_warning(TLS, "warning: gnutls_certificate_set_x509_system_trust() failed: (%d) %s\n",
 				     err, gnutls_strerror_name(err));
 			tls_credentials_free(tls_credentials);
 			return err;
@@ -709,7 +708,7 @@ int tls_certificate_set(struct network *net, const char *tls_cert, const char *t
 	if ((err = gnutls_certificate_set_x509_key_file(tls_credentials->credentials,
 							tls_cert, tls_key, GNUTLS_X509_FMT_PEM)) != GNUTLS_E_SUCCESS) {
 		tls_credentials_free(tls_credentials);
-		kr_log_error("[tls] gnutls_certificate_set_x509_key_file(%s,%s) failed: %d (%s)\n",
+		kr_log_error(TLS, "gnutls_certificate_set_x509_key_file(%s,%s) failed: %d (%s)\n",
 			     tls_cert, tls_key, err, gnutls_strerror_name(err));
 		return kr_error(EINVAL);
 	}
@@ -775,8 +774,7 @@ void tls_credentials_free(struct tls_credentials *tls_credentials) {
 
 void tls_client_param_unref(tls_client_param_t *entry)
 {
-	if (!entry) return;
-	assert(entry->refs); /* Well, we'd only leak memory. */
+	if (!entry || kr_fails_assert(entry->refs)) return;
 	--(entry->refs);
 	if (entry->refs) return;
 
@@ -806,7 +804,8 @@ void tls_client_param_unref(tls_client_param_t *entry)
 }
 static int param_free(void **param, void *null)
 {
-	assert(param && *param);
+	if (kr_fails_assert(param && *param))
+		return -1;
 	tls_client_param_unref(*param);
 	return 0;
 }
@@ -820,15 +819,13 @@ void tls_client_params_free(tls_client_params_t *params)
 tls_client_param_t * tls_client_param_new()
 {
 	tls_client_param_t *e = calloc(1, sizeof(*e));
-	if (!e) {
-		assert(!ENOMEM);
+	if (kr_fails_assert(e))
 		return NULL;
-	}
 	/* Note: those array_t don't need further initialization. */
 	e->refs = 1;
 	int ret = gnutls_certificate_allocate_credentials(&e->credentials);
 	if (ret != GNUTLS_E_SUCCESS) {
-		kr_log_error("[tls_client] error: gnutls_certificate_allocate_credentials() fails (%s)\n",
+		kr_log_error(TLSCLIENT, "error: gnutls_certificate_allocate_credentials() fails (%s)\n",
 			     gnutls_strerror_name(ret));
 		free(e);
 		return NULL;
@@ -861,22 +858,21 @@ static bool construct_key(const union inaddr *addr, uint32_t *len, char *key)
 		*len = sizeof(addr->ip6.sin6_port) + sizeof(addr->ip6.sin6_addr);
 		return true;
 	default:
-		assert(!EINVAL);
+		kr_assert(!EINVAL);
 		return false;
 	}
 }
 tls_client_param_t ** tls_client_param_getptr(tls_client_params_t **params,
 				const struct sockaddr *addr, bool do_insert)
 {
-	assert(params && addr);
+	if (kr_fails_assert(params && addr))
+		return NULL;
 	/* We accept NULL for empty map; ensure the map exists if needed. */
 	if (!*params) {
 		if (!do_insert) return NULL;
 		*params = trie_create(NULL);
-		if (!*params) {
-			assert(!ENOMEM);
+		if (kr_fails_assert(*params))
 			return NULL;
-		}
 	}
 	/* Construct the key. */
 	const union inaddr *ia = (const union inaddr *)addr;
@@ -913,7 +909,8 @@ static int client_verify_pin(const unsigned int cert_list_size,
 				const gnutls_datum_t *cert_list,
 				tls_client_param_t *params)
 {
-	assert(params->pins.len > 0);
+	if (kr_fails_assert(params->pins.len > 0))
+		return GNUTLS_E_CERTIFICATE_ERROR;
 #if TLS_CAN_USE_PINS
 	for (int i = 0; i < cert_list_size; i++) {
 		gnutls_x509_crt_t cert;
@@ -929,7 +926,7 @@ static int client_verify_pin(const unsigned int cert_list_size,
 		}
 
 	#ifdef DEBUG
-		if (VERBOSE_STATUS) {
+		if (kr_log_is_debug(TLS, NULL)) {
 			char pin_base64[TLS_SHA256_BASE64_BUFLEN];
 			/* DEBUG: additionally compute and print the base64 pin.
 			 * Not very efficient, but that's OK for DEBUG. */
@@ -960,13 +957,13 @@ static int client_verify_pin(const unsigned int cert_list_size,
 				params->pins.len);
 	}
 
-	kr_log_error("[tls_client] no pin matched: %zu pins * %d certificates\n",
+	kr_log_error(TLSCLIENT, "no pin matched: %zu pins * %d certificates\n",
 			params->pins.len, cert_list_size);
 	return GNUTLS_E_CERTIFICATE_ERROR;
 
 #else /* TLS_CAN_USE_PINS */
-	kr_log_error("[tls_client] internal inconsistency: TLS_CAN_USE_PINS\n");
-	assert(false);
+	kr_log_error(TLSCLIENT, "internal inconsistency: TLS_CAN_USE_PINS\n");
+	kr_assert(false);
 	return GNUTLS_E_CERTIFICATE_ERROR;
 #endif
 }
@@ -979,9 +976,8 @@ static int client_verify_pin(const unsigned int cert_list_size,
  */
 static int client_verify_certchain(gnutls_session_t tls_session, const char *hostname)
 {
-	if (!hostname) {
-		kr_log_error("[tls_client] internal config inconsistency: no hostname set\n");
-		assert(false);
+	if (kr_fails_assert(hostname)) {
+		kr_log_error(TLSCLIENT, "internal config inconsistency: no hostname set\n");
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
@@ -996,16 +992,16 @@ static int client_verify_certchain(gnutls_session_t tls_session, const char *hos
 		ret = gnutls_certificate_verification_status_print(
 			status, gnutls_certificate_type_get(tls_session), &msg, 0);
 		if (ret == GNUTLS_E_SUCCESS) {
-			kr_log_error("[tls_client] failed to verify peer certificate: "
+			kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
 					"%s\n", msg.data);
 			gnutls_free(msg.data);
 		} else {
-			kr_log_error("[tls_client] failed to verify peer certificate: "
+			kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
 					"unable to print reason: %s (%s)\n",
 					gnutls_strerror(ret), gnutls_strerror_name(ret));
 		} /* gnutls_certificate_verification_status_print end */
 	} else {
-		kr_log_error("[tls_client] failed to verify peer certificate: "
+		kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
 			     "gnutls_certificate_verify_peers3 error: %s (%s)\n",
 			     gnutls_strerror(ret), gnutls_strerror_name(ret));
 	} /* gnutls_certificate_verify_peers3 end */
@@ -1020,7 +1016,8 @@ static int client_verify_certchain(gnutls_session_t tls_session, const char *hos
 static int client_verify_certificate(gnutls_session_t tls_session)
 {
 	struct tls_client_ctx *ctx = gnutls_session_get_ptr(tls_session);
-	assert(ctx->params != NULL);
+	if (kr_fails_assert(ctx->params))
+		return GNUTLS_E_CERTIFICATE_ERROR;
 
 	if (ctx->params->insecure) {
 		return GNUTLS_E_SUCCESS;
@@ -1028,7 +1025,7 @@ static int client_verify_certificate(gnutls_session_t tls_session)
 
 	gnutls_certificate_type_t cert_type = gnutls_certificate_type_get(tls_session);
 	if (cert_type != GNUTLS_CRT_X509) {
-		kr_log_error("[tls_client] invalid certificate type %i has been received\n",
+		kr_log_error(TLSCLIENT, "invalid certificate type %i has been received\n",
 			     cert_type);
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
@@ -1036,7 +1033,7 @@ static int client_verify_certificate(gnutls_session_t tls_session)
 	const gnutls_datum_t *cert_list =
 		gnutls_certificate_get_peers(tls_session, &cert_list_size);
 	if (cert_list == NULL || cert_list_size == 0) {
-		kr_log_error("[tls_client] empty certificate list\n");
+		kr_log_error(TLSCLIENT, "empty certificate list\n");
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
@@ -1081,9 +1078,9 @@ struct tls_client_ctx *tls_client_ctx_new(tls_client_param_t *entry,
 	if (ret == GNUTLS_E_SUCCESS && entry->hostname) {
 		ret = gnutls_server_name_set(ctx->c.tls_session, GNUTLS_NAME_DNS,
 					entry->hostname, strlen(entry->hostname));
-		kr_log_verbose("[tls_client] set hostname, ret = %d\n", ret);
+		kr_log_debug(TLSCLIENT, "set hostname, ret = %d\n", ret);
 	} else if (!entry->hostname) {
-		kr_log_verbose("[tls_client] no hostname\n");
+		kr_log_debug(TLSCLIENT, "no hostname\n");
 	}
 	if (ret != GNUTLS_E_SUCCESS) {
 		tls_client_ctx_free(ctx);
@@ -1119,8 +1116,11 @@ void tls_client_ctx_free(struct tls_client_ctx *ctx)
 int  tls_pull_timeout_func(gnutls_transport_ptr_t h, unsigned int ms)
 {
 	struct tls_common_ctx *t = (struct tls_common_ctx *)h;
-	assert(t != NULL);
-	ssize_t	avail = t->nread - t->consumed;
+	if (kr_fails_assert(t)) {
+		errno = EFAULT;
+		return -1;
+	}
+	ssize_t avail = t->nread - t->consumed;
 	DEBUG_MSG("[%s] timeout check: available: %zu\n",
 		  t->client_side ? "tls_client" : "tls", avail);
 	if (avail <= 0) {
@@ -1134,11 +1134,11 @@ int tls_client_connect_start(struct tls_client_ctx *client_ctx,
 			     struct session *session,
 			     tls_handshake_cb handshake_cb)
 {
-	if (session == NULL || client_ctx == NULL) {
+	if (session == NULL || client_ctx == NULL)
 		return kr_error(EINVAL);
-	}
 
-	assert(session_flags(session)->outgoing && session_get_handle(session)->type == UV_TCP);
+	if (kr_fails_assert(session_flags(session)->outgoing && session_get_handle(session)->type == UV_TCP))
+		return kr_error(EINVAL);
 
 	struct tls_common_ctx *ctx = &client_ctx->c;
 
@@ -1190,3 +1190,4 @@ int tls_client_ctx_set_session(struct tls_client_ctx *ctx, struct session *sessi
 }
 
 #undef DEBUG_MSG
+#undef VERBOSE_MSG
