@@ -16,7 +16,6 @@
 #if defined(__GLIBC__) && defined(_GNU_SOURCE)
 #include <malloc.h>
 #endif
-#include <assert.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <gnutls/gnutls.h>
@@ -52,7 +51,7 @@
 #define MAX_PIPELINED 100
 #endif
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "wrkr", __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, WORKER, __VA_ARGS__)
 
 /** Client request state. */
 struct request_ctx
@@ -100,7 +99,7 @@ struct qr_task
 #define qr_task_unref(task) \
 	do { \
 		if (task) \
-			assert((task)->refs > 0); \
+			kr_require((task)->refs > 0); \
 		if ((task) && --(task)->refs == 0) \
 			qr_task_free((task)); \
 	} while (0)
@@ -142,9 +141,8 @@ static uv_handle_t *ioreq_spawn(struct worker_ctx *worker,
 {
 	bool precond = (socktype == SOCK_DGRAM || socktype == SOCK_STREAM)
 			&& (family == AF_INET  || family == AF_INET6);
-	if (!precond) {
-		assert(false);
-		kr_log_verbose("[work] ioreq_spawn: pre-condition failed\n");
+	if (kr_fails_assert(precond)) {
+		kr_log_debug(WORKER, "ioreq_spawn: pre-condition failed\n");
 		return NULL;
 	}
 
@@ -172,7 +170,10 @@ static uv_handle_t *ioreq_spawn(struct worker_ctx *worker,
 		addr = (union inaddr *)&worker->out_addr6;
 	}
 	if (addr->ip.sa_family != AF_UNSPEC) {
-		assert(addr->ip.sa_family == family);
+		if (kr_fails_assert(addr->ip.sa_family == family)) {
+			io_free(handle);
+			return NULL;
+		}
 		if (socktype == SOCK_DGRAM) {
 			uv_udp_t *udp = (uv_udp_t *)handle;
 			ret = uv_udp_bind(udp, &addr->ip, 0);
@@ -259,7 +260,7 @@ static inline void pool_release(struct worker_ctx *worker, struct mempool *mp)
 static const size_t SUBREQ_KEY_LEN = KR_RRKEY_LEN;
 static int subreq_key(char *dst, knot_pkt_t *pkt)
 {
-	assert(pkt);
+	kr_require(pkt);
 	return kr_rrkey(dst, knot_pkt_qclass(pkt), knot_pkt_qname(pkt),
 			knot_pkt_qtype(pkt), knot_pkt_qtype(pkt));
 }
@@ -267,11 +268,13 @@ static int subreq_key(char *dst, knot_pkt_t *pkt)
 #if ENABLE_XDP
 static uint8_t *alloc_wire_cb(struct kr_request *req, uint16_t *maxlen)
 {
-	assert(maxlen);
+	if (kr_fails_assert(maxlen))
+		return NULL;
 	struct request_ctx *ctx = (struct request_ctx *)req;
 	/* We know it's an AF_XDP socket; otherwise this CB isn't assigned. */
 	uv_handle_t *handle = session_get_handle(ctx->source.session);
-	assert(handle->type == UV_POLL);
+	if (kr_fails_assert(handle->type == UV_POLL))
+		return NULL;
 	xdp_handle_data_t *xhd = handle->data;
 	knot_xdp_msg_t out;
 	bool ipv6 = ctx->source.addr.ip.sa_family == AF_INET6;
@@ -282,7 +285,7 @@ static uint8_t *alloc_wire_cb(struct kr_request *req, uint16_t *maxlen)
 					ipv6, &out, NULL);
 			#endif
 	if (ret != KNOT_EOK) {
-		assert(ret == KNOT_ENOMEM);
+		kr_assert(ret == KNOT_ENOMEM);
 		*maxlen = 0;
 		return NULL;
 	}
@@ -294,7 +297,8 @@ static uint8_t *alloc_wire_cb(struct kr_request *req, uint16_t *maxlen)
 }
 static void free_wire(const struct request_ctx *ctx)
 {
-	assert(ctx->req.alloc_wire_cb == alloc_wire_cb);
+	if (kr_fails_assert(ctx->req.alloc_wire_cb == alloc_wire_cb))
+		return;
 	knot_pkt_t *ans = ctx->req.answer;
 	if (unlikely(ans == NULL)) /* dropped */
 		return;
@@ -302,7 +306,8 @@ static void free_wire(const struct request_ctx *ctx)
 		return;
 	/* We know it's an AF_XDP socket; otherwise alloc_wire_cb isn't assigned. */
 	uv_handle_t *handle = session_get_handle(ctx->source.session);
-	assert(handle->type == UV_POLL);
+	if (kr_fails_assert(handle->type == UV_POLL))
+		return;
 	xdp_handle_data_t *xhd = handle->data;
 	/* Freeing is done by sending an empty packet (the API won't really send it). */
 	knot_xdp_msg_t out;
@@ -310,8 +315,8 @@ static void free_wire(const struct request_ctx *ctx)
 	out.payload.iov_len = 0;
 	uint32_t sent;
 	int ret = knot_xdp_send(xhd->socket, &out, 1, &sent);
-	assert(ret == KNOT_EOK && sent == 0); (void)ret;
-	kr_log_verbose("[xdp] freed unsent buffer, ret = %d\n", ret);
+	kr_assert(ret == KNOT_EOK && sent == 0);
+	kr_log_debug(XDP, "freed unsent buffer, ret = %d\n", ret);
 }
 #endif
 /* Helper functions for transport selection */
@@ -355,20 +360,28 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 
 	/* TODO Relocate pool to struct request */
 	ctx->worker = worker;
-	if (session) {
-		assert(session_flags(session)->outgoing == false);
+	if (session && kr_fails_assert(session_flags(session)->outgoing == false)) {
+		pool_release(worker, pool.ctx);
+		return NULL;
 	}
 	ctx->source.session = session;
-	assert(!!eth_to == !!eth_from);
+	if (kr_fails_assert(!!eth_to == !!eth_from)) {
+		pool_release(worker, pool.ctx);
+		return NULL;
+	}
 	const bool is_xdp = eth_to != NULL;
 	if (is_xdp) {
 	#if ENABLE_XDP
-		assert(session);
+		if (kr_fails_assert(session)) {
+			pool_release(worker, pool.ctx);
+			return NULL;
+		}
 		memcpy(&ctx->source.eth_addrs[0], eth_to,   sizeof(ctx->source.eth_addrs[0]));
 		memcpy(&ctx->source.eth_addrs[1], eth_from, sizeof(ctx->source.eth_addrs[1]));
 		ctx->req.alloc_wire_cb = alloc_wire_cb;
 	#else
-		assert(!EINVAL);
+		kr_assert(!EINVAL);
+		pool_release(worker, pool.ctx);
 		return NULL;
 	#endif
 	}
@@ -378,6 +391,7 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 	req->vars_ref = LUA_NOREF;
 	req->uid = uid;
 	req->qsource.flags.xdp = is_xdp;
+	array_init(req->qsource.headers);
 	if (session) {
 		req->qsource.flags.tcp = session_get_handle(session)->type == UV_TCP;
 		req->qsource.flags.tls = session_flags(session)->has_tls;
@@ -386,7 +400,13 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 #if ENABLE_DOH2
 		if (req->qsource.flags.http) {
 			struct http_ctx *http_ctx = session_http_get_server_ctx(session);
-			req->qsource.stream_id = queue_head(http_ctx->streams);
+			struct http_stream stream = queue_head(http_ctx->streams);
+			req->qsource.stream_id = stream.id;
+			if (stream.headers) {
+				req->qsource.headers = *stream.headers;
+				free(stream.headers);
+				stream.headers = NULL;
+			}
 		}
 #endif
 		/* We need to store a copy of peer address. */
@@ -412,7 +432,8 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 /** More initialization, related to the particular incoming query/packet. */
 static int request_start(struct request_ctx *ctx, knot_pkt_t *query)
 {
-	assert(query && ctx);
+	if (kr_fails_assert(query && ctx))
+		return kr_error(EINVAL);
 
 	struct kr_request *req = &ctx->req;
 	req->qsource.size = query->size;
@@ -456,12 +477,19 @@ static void request_free(struct request_ctx *ctx)
 		lua_pop(L, 1);
 		ctx->req.vars_ref = LUA_NOREF;
 	}
+	/* Free HTTP/2 headers for DoH requests. */
+	for(int i = 0; i < ctx->req.qsource.headers.len; i++) {
+		free(ctx->req.qsource.headers.at[i].name);
+		free(ctx->req.qsource.headers.at[i].value);
+	}
+	array_clear(ctx->req.qsource.headers);
+
 	/* Make sure to free XDP buffer in case it wasn't sent. */
 	if (ctx->req.alloc_wire_cb) {
 	#if ENABLE_XDP
 		free_wire(ctx);
 	#else
-		assert(!EINVAL);
+		kr_assert(!EINVAL);
 	#endif
 	}
 	/* Return mempool to ring or free it if it's full */
@@ -499,7 +527,7 @@ static struct qr_task *qr_task_create(struct request_ctx *ctx)
 	task->pktbuf = pktbuf;
 	array_init(task->waiting);
 	task->refs = 0;
-	assert(ctx->task == NULL);
+	kr_assert(ctx->task == NULL);
 	ctx->task = task;
 	/* Make the primary reference to task. */
 	qr_task_ref(task);
@@ -513,7 +541,8 @@ static void qr_task_free(struct qr_task *task)
 {
 	struct request_ctx *ctx = task->ctx;
 
-	assert(ctx);
+	if (kr_fails_assert(ctx))
+		return;
 
 	struct worker_ctx *worker = ctx->worker;
 
@@ -528,12 +557,14 @@ static void qr_task_free(struct qr_task *task)
 /*@ Register new qr_task within session. */
 static int qr_task_register(struct qr_task *task, struct session *session)
 {
-	assert(!session_flags(session)->outgoing && session_get_handle(session)->type == UV_TCP);
+	if (kr_fails_assert(!session_flags(session)->outgoing && session_get_handle(session)->type == UV_TCP))
+		return kr_error(EINVAL);
 
 	session_tasklist_add(session, task);
 
 	struct request_ctx *ctx = task->ctx;
-	assert(ctx && (ctx->source.session == NULL || ctx->source.session == session));
+	if (kr_fails_assert(ctx && (ctx->source.session == NULL || ctx->source.session == session)))
+		return kr_error(EINVAL);
 	ctx->source.session = session;
 	/* Soft-limit on parallel queries, there is no "slow down" RCODE
 	 * that we could use to signalize to client, but we can stop reading,
@@ -555,12 +586,12 @@ static void qr_task_complete(struct qr_task *task)
 
 	/* Kill pending I/O requests */
 	ioreq_kill_pending(task);
-	assert(task->waiting.len == 0);
-	assert(task->leading == false);
+	kr_require(task->waiting.len == 0);
+	kr_require(task->leading == false);
 
 	struct session *s = ctx->source.session;
 	if (s) {
-		assert(!session_flags(s)->outgoing && session_waitinglist_is_empty(s));
+		kr_require(!session_flags(s)->outgoing && session_waitinglist_is_empty(s));
 		ctx->source.session = NULL;
 		session_tasklist_del(s, task);
 	}
@@ -576,23 +607,22 @@ static void qr_task_complete(struct qr_task *task)
 int qr_task_on_send(struct qr_task *task, const uv_handle_t *handle, int status)
 {
 	if (task->finished) {
-		assert(task->leading == false);
+		kr_require(task->leading == false);
 		qr_task_complete(task);
 	}
 
-	if (!handle)
+	if (!handle || kr_fails_assert(handle->data))
 		return status;
-
 	struct session* s = handle->data;
-	assert(s);
 
 	if (handle->type == UV_UDP && session_flags(s)->outgoing) {
 		// This should ensure that we are only dealing with our question to upstream
-		assert(!knot_wire_get_qr(task->pktbuf->wire));
+		if (kr_fails_assert(!knot_wire_get_qr(task->pktbuf->wire)))
+			return status;
 		// start the timer
 		struct kr_query *qry = array_tail(task->ctx->req.rplan.pending);
-		assert(qry != NULL);
-		(void)qry;
+		if (kr_fails_assert(qry))
+			return status;
 		size_t timeout = task->transport->timeout;
 		int ret = session_timer_start(s, on_udp_timeout, timeout, 0);
 		/* Start next step with timeout, fatal if can't start a timer. */
@@ -603,8 +633,24 @@ int qr_task_on_send(struct qr_task *task, const uv_handle_t *handle, int status)
 	}
 
 	if (handle->type == UV_TCP) {
-		if (status != 0)
-			session_tasklist_del(s, task);
+		if (status != 0) { // session probably not usable anymore; typically: ECONNRESET
+			const struct kr_request *req = &task->ctx->req;
+			if (kr_log_is_debug(WORKER, req)) {
+				const char *peer_str = NULL;
+				if (!session_flags(s)->outgoing) {
+					peer_str = "hidden"; // avoid logging downstream IPs
+				} else if (task->transport) {
+					peer_str = kr_straddr(&task->transport->address.ip);
+				}
+				if (!peer_str)
+					peer_str = "unknown"; // probably shouldn't happen
+				kr_log_req(req, 0, 0, WORKER,
+						"=> disconnected from '%s': %s\n",
+						peer_str, uv_strerror(status));
+			}
+			worker_end_tcp(s);
+			return status;
+		}
 
 		if (session_flags(s)->outgoing || session_flags(s)->closing)
 			return status;
@@ -643,25 +689,23 @@ static void on_write(uv_write_t *req, int status)
 static int qr_task_send(struct qr_task *task, struct session *session,
 			const struct sockaddr *addr, knot_pkt_t *pkt)
 {
-	if (!session) {
+	if (!session)
 		return qr_task_on_send(task, NULL, kr_error(EIO));
-	}
 
 	int ret = 0;
 	struct request_ctx *ctx = task->ctx;
 
 	uv_handle_t *handle = session_get_handle(session);
-	assert(handle && handle->data == session);
+	if (kr_fails_assert(handle && handle->data == session))
+		return qr_task_on_send(task, NULL, kr_error(EINVAL));
 	const bool is_stream = handle->type == UV_TCP;
 	if (!is_stream && handle->type != UV_UDP) abort();
 
-	if (addr == NULL) {
+	if (addr == NULL)
 		addr = session_get_peer(session);
-	}
 
-	if (pkt == NULL) {
+	if (pkt == NULL)
 		pkt = worker_task_get_pktbuf(task);
-	}
 
 	if (session_flags(session)->outgoing && handle->type == UV_TCP) {
 		size_t try_limit = session_tasklist_get_len(session) + 1;
@@ -672,16 +716,14 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 			++msg_id;
 			++try_count;
 		}
-		if (try_count > try_limit) {
+		if (try_count > try_limit)
 			return kr_error(ENOENT);
-		}
 		worker_task_pkt_set_msgid(task, msg_id);
 	}
 
 	uv_handle_t *ioreq = malloc(is_stream ? sizeof(uv_write_t) : sizeof(uv_udp_send_t));
-	if (!ioreq) {
+	if (!ioreq)
 		return qr_task_on_send(task, handle, kr_error(ENOMEM));
-	}
 
 	/* Pending ioreq on current task */
 	qr_task_ref(task);
@@ -691,7 +733,8 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 	task->send_time = kr_now();
 	task->recv_time = 0; // task structure is being reused so we have to zero this out here
 	/* Send using given protocol */
-	assert(!session_flags(session)->closing);
+	if (kr_fails_assert(!session_flags(session)->closing))
+		return qr_task_on_send(task, NULL, kr_error(EIO));
 	if (session_flags(session)->has_http) {
 #if ENABLE_DOH2
 		uv_write_t *write_req = (uv_write_t *)ioreq;
@@ -734,7 +777,7 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 		write_req->data = task;
 		ret = uv_write(write_req, (uv_stream_t *)handle, buf, 3, &on_write);
 	} else {
-		assert(false);
+		kr_assert(false);
 	}
 
 	if (ret == 0) {
@@ -794,7 +837,8 @@ static struct kr_query *task_get_last_pending_query(struct qr_task *task)
 
 static int session_tls_hs_cb(struct session *session, int status)
 {
-	assert(session_flags(session)->outgoing);
+	if (kr_fails_assert(session_flags(session)->outgoing))
+		return kr_error(EINVAL);
 	struct sockaddr *peer = session_get_peer(session);
 	int deletion_res = worker_del_tcp_waiting(the_worker, peer);
 	int ret = kr_ok();
@@ -812,10 +856,10 @@ static int session_tls_hs_cb(struct session *session, int status)
 			 * waiting for connection to upstream.
 			 * So that it MUST be unsuccessful rehandshake.
 			 * Check it. */
-			assert(deletion_res != 0);
+			kr_require(deletion_res != 0);
 			const char *key = tcpsess_key(peer);
-			assert(key);
-			assert(map_contains(&the_worker->tcp_connected, key) != 0);
+			kr_require(key);
+			kr_require(map_contains(&the_worker->tcp_connected, key) != 0);
 		}
 #endif
 		return ret;
@@ -826,9 +870,9 @@ static int session_tls_hs_cb(struct session *session, int status)
 	tls_client_param_t *tls_params = tls_client_ctx->params;
 	gnutls_session_t tls_session = tls_client_ctx->c.tls_session;
 	if (gnutls_session_is_resumed(tls_session) != 0) {
-		kr_log_verbose("[tls_client] TLS session has resumed\n");
+		kr_log_debug(TLSCLIENT, "TLS session has resumed\n");
 	} else {
-		kr_log_verbose("[tls_client] TLS session has not resumed\n");
+		kr_log_debug(TLSCLIENT, "TLS session has not resumed\n");
 		/* session wasn't resumed, delete old session data ... */
 		if (tls_params->session_data.data != NULL) {
 			gnutls_free(tls_params->session_data.data);
@@ -919,20 +963,23 @@ static int send_waiting(struct session *session)
 static void on_connect(uv_connect_t *req, int status)
 {
 	struct worker_ctx *worker = the_worker;
-	assert(worker);
+	kr_require(worker);
 	uv_stream_t *handle = req->handle;
 	struct session *session = handle->data;
 	struct sockaddr *peer = session_get_peer(session);
 	free(req);
 
-	assert(session_flags(session)->outgoing);
+	if (kr_fails_assert(session_flags(session)->outgoing))
+		return;
 
 	if (session_flags(session)->closing) {
 		worker_del_tcp_waiting(worker, peer);
-		assert(session_is_empty(session));
+		kr_assert(session_is_empty(session));
 		return;
 	}
 
+	const bool log_debug = kr_log_is_debug(WORKER, NULL);
+
 	/* Check if the connection is in the waiting list.
 	 * If no, most likely this is timeouted connection
 	 * which was removed from waiting list by
@@ -941,13 +988,13 @@ static void on_connect(uv_connect_t *req, int status)
 	if (!s || s != session) {
 		/* session isn't on the waiting list.
 		 * it's timeouted session. */
-		if (VERBOSE_STATUS) {
+		if (log_debug) {
 			const char *peer_str = kr_straddr(peer);
-			kr_log_verbose( "[wrkr]=> connected to '%s', but session "
+			kr_log_debug(WORKER, "=> connected to '%s', but session "
 					"is already timeouted, close\n",
 					peer_str ? peer_str : "");
 		}
-		assert(session_tasklist_is_empty(session));
+		kr_assert(session_tasklist_is_empty(session));
 		session_waitinglist_retry(session, false);
 		session_close(session);
 		return;
@@ -958,22 +1005,22 @@ static void on_connect(uv_connect_t *req, int status)
 		/* session already in the connected list.
 		 * Something went wrong, it can be due to races when kresd has tried
 		 * to reconnect to upstream after unsuccessful attempt. */
-		if (VERBOSE_STATUS) {
+		if (log_debug) {
 			const char *peer_str = kr_straddr(peer);
-			kr_log_verbose( "[wrkr]=> connected to '%s', but peer "
+			kr_log_debug(WORKER, "=> connected to '%s', but peer "
 					"is already connected, close\n",
 					peer_str ? peer_str : "");
 		}
-		assert(session_tasklist_is_empty(session));
+		kr_assert(session_tasklist_is_empty(session));
 		session_waitinglist_retry(session, false);
 		session_close(session);
 		return;
 	}
 
 	if (status != 0) {
-		if (VERBOSE_STATUS) {
+		if (log_debug) {
 			const char *peer_str = kr_straddr(peer);
-			kr_log_verbose( "[wrkr]=> connection to '%s' failed (%s), flagged as 'bad'\n",
+			kr_log_debug(WORKER, "=> connection to '%s' failed (%s), flagged as 'bad'\n",
 					peer_str ? peer_str : "", uv_strerror(status));
 		}
 		worker_del_tcp_waiting(worker, peer);
@@ -985,7 +1032,7 @@ static void on_connect(uv_connect_t *req, int status)
 			struct kr_query *qry = array_tail(task->ctx->req.rplan.pending);
 			qry->server_selection.error(qry, task->transport, KR_SELECTION_TCP_CONNECT_FAILED);
 		}
-		assert(session_tasklist_is_empty(session));
+		kr_assert(session_tasklist_is_empty(session));
 		session_waitinglist_retry(session, false);
 		session_close(session);
 		return;
@@ -998,15 +1045,15 @@ static void on_connect(uv_connect_t *req, int status)
 			/* session isn't in list of waiting queries, *
 			 * something gone wrong */
 			session_waitinglist_finalize(session, KR_STATE_FAIL);
-			assert(session_tasklist_is_empty(session));
+			kr_assert(session_tasklist_is_empty(session));
 			session_close(session);
 			return;
 		}
 	}
 
-	if (VERBOSE_STATUS) {
+	if (log_debug) {
 		const char *peer_str = kr_straddr(peer);
-		kr_log_verbose( "[wrkr]=> connected to '%s'\n", peer_str ? peer_str : "");
+		kr_log_debug(WORKER, "=> connected to '%s'\n", peer_str ? peer_str : "");
 	}
 
 	session_flags(session)->connected = true;
@@ -1042,9 +1089,9 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 
 	uv_timer_stop(timer);
 	struct worker_ctx *worker = the_worker;
-	assert(worker);
+	kr_require(worker);
 
-	assert (session_tasklist_is_empty(session));
+	kr_assert(session_tasklist_is_empty(session));
 
 	struct sockaddr *peer = session_get_peer(session);
 	worker_del_tcp_waiting(worker, peer);
@@ -1059,7 +1106,7 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 	}
 
 	struct kr_query *qry = task_get_last_pending_query(task);
-	WITH_VERBOSE (qry) {
+	if (kr_log_is_debug_qry(WORKER, qry)) {
 		const char *peer_str = kr_straddr(peer);
 		VERBOSE_MSG(qry, "=> connection to '%s' failed (internal timeout)\n",
 			    peer_str ? peer_str : "");
@@ -1069,7 +1116,7 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 
 	worker->stats.timeout += session_waitinglist_get_len(session);
 	session_waitinglist_retry(session, true);
-	assert (session_tasklist_is_empty(session));
+	kr_assert(session_tasklist_is_empty(session));
 	/* uv_cancel() doesn't support uv_connect_t request,
 	 * so that we can't cancel it.
 	 * There still exists possibility of successful connection
@@ -1084,13 +1131,15 @@ static void on_tcp_connect_timeout(uv_timer_t *timer)
 static void on_udp_timeout(uv_timer_t *timer)
 {
 	struct session *session = timer->data;
-	assert(session_get_handle(session)->data == session);
-	assert(session_tasklist_get_len(session) == 1);
-	assert(session_waitinglist_is_empty(session));
+	kr_assert(session_get_handle(session)->data == session);
+	kr_assert(session_tasklist_get_len(session) == 1);
+	kr_assert(session_waitinglist_is_empty(session));
 
 	uv_timer_stop(timer);
 
 	struct qr_task *task = session_tasklist_get_first(session);
+	if (!task)
+		return;
 	struct worker_ctx *worker = task->ctx->worker;
 
 	if (task->leading && task->pending_count > 0) {
@@ -1130,7 +1179,7 @@ static uv_handle_t *transmit(struct qr_task *task)
 		struct sockaddr *addr = (struct sockaddr *)choice;
 		struct session *session = ret->data;
 		struct sockaddr *peer = session_get_peer(session);
-		assert (peer->sa_family == AF_UNSPEC && session_flags(session)->outgoing);
+		kr_assert(peer->sa_family == AF_UNSPEC && session_flags(session)->outgoing);
 		memcpy(peer, addr, kr_sockaddr_len(addr));
 		if (qr_task_send(task, session, (struct sockaddr *)choice,
 				 task->pktbuf) != 0) {
@@ -1161,7 +1210,7 @@ static void subreq_finalize(struct qr_task *task, const struct sockaddr *packet_
 	if (klen > 0) {
 		void *val_deleted;
 		int ret = trie_del(task->ctx->worker->subreq_out, key, klen, &val_deleted);
-		assert(ret == KNOT_EOK && val_deleted == task); (void)ret;
+		kr_assert(ret == KNOT_EOK && val_deleted == task);
 	}
 	/* Notify waiting tasks. */
 	struct kr_query *leader_qry = array_tail(task->ctx->req.rplan.pending);
@@ -1189,7 +1238,8 @@ static void subreq_finalize(struct qr_task *task, const struct sockaddr *packet_
 
 static void subreq_lead(struct qr_task *task)
 {
-	assert(task);
+	if (kr_fails_assert(task))
+		return;
 	char key[SUBREQ_KEY_LEN];
 	const int klen = subreq_key(key, task->pktbuf);
 	if (klen < 0)
@@ -1198,17 +1248,16 @@ static void subreq_lead(struct qr_task *task)
 		trie_get_ins(task->ctx->worker->subreq_out, key, klen);
 	if (unlikely(!tvp))
 		return; /*ENOMEM*/
-	if (unlikely(*tvp != NULL)) {
-		assert(false);
+	if (kr_fails_assert(*tvp == NULL))
 		return;
-	}
 	*tvp = task;
 	task->leading = true;
 }
 
 static bool subreq_enqueue(struct qr_task *task)
 {
-	assert(task);
+	if (kr_fails_assert(task))
+		return false;
 	char key[SUBREQ_KEY_LEN];
 	const int klen = subreq_key(key, task->pktbuf);
 	if (klen < 0)
@@ -1231,7 +1280,7 @@ static void xdp_tx_waker(uv_idle_t *handle)
 {
 	int ret = knot_xdp_send_finish(handle->data);
 	if (ret != KNOT_EAGAIN && ret != KNOT_EOK)
-		kr_log_error("[xdp] check: ret = %d, %s\n", ret, knot_strerror(ret));
+		kr_log_error(XDP, "check: ret = %d, %s\n", ret, knot_strerror(ret));
 	/* Apparently some drivers need many explicit wake-up calls
 	 * even if we push no additional packets (in case they accumulated a lot) */
 	if (ret != KNOT_EAGAIN)
@@ -1246,6 +1295,10 @@ static int xdp_push(struct qr_task *task, const uv_handle_t *src_handle)
 {
 #if ENABLE_XDP
 	struct request_ctx *ctx = task->ctx;
+	xdp_handle_data_t *xhd = src_handle->data;
+	if (kr_fails_assert(xhd && xhd->socket && xhd->session == ctx->source.session))
+		return qr_task_on_send(task, src_handle, kr_error(EINVAL));
+
 	knot_xdp_msg_t msg;
 	const struct sockaddr *ip_from = &ctx->source.dst_addr.ip;
 	const struct sockaddr *ip_to   = &ctx->source.addr.ip;
@@ -1254,25 +1307,23 @@ static int xdp_push(struct qr_task *task, const uv_handle_t *src_handle)
 	msg.payload.iov_base = ctx->req.answer->wire;
 	msg.payload.iov_len  = ctx->req.answer->size;
 
-	xdp_handle_data_t *xhd = src_handle->data;
-	assert(xhd && xhd->socket && xhd->session == ctx->source.session);
 	uint32_t sent;
 	int ret = knot_xdp_send(xhd->socket, &msg, 1, &sent);
 	ctx->req.answer->wire = NULL; /* it's been freed */
 
 	uv_idle_start(&xhd->tx_waker, xdp_tx_waker);
-	kr_log_verbose("[xdp] pushed a packet, ret = %d\n", ret);
+	kr_log_debug(XDP, "pushed a packet, ret = %d\n", ret);
 
 	return qr_task_on_send(task, src_handle, ret);
 #else
-	assert(!EINVAL);
+	kr_assert(!EINVAL);
 	return kr_error(EINVAL);
 #endif
 }
 
 static int qr_task_finalize(struct qr_task *task, int state)
 {
-	assert(task && task->leading == false);
+	kr_require(task && task->leading == false);
 	if (task->finished) {
 		return kr_ok();
 	}
@@ -1301,21 +1352,18 @@ static int qr_task_finalize(struct qr_task *task, int state)
 	/* Send back answer */
 	int ret;
 	const uv_handle_t *src_handle = session_get_handle(source_session);
-	if (src_handle->type != UV_UDP && src_handle->type != UV_TCP
-				       && src_handle->type != UV_POLL) {
-		assert(false);
+	if (kr_fails_assert(src_handle->type == UV_UDP || src_handle->type == UV_TCP
+		       || src_handle->type == UV_POLL)) {
 		ret = kr_error(EINVAL);
-
 	} else if (src_handle->type == UV_POLL) {
 		ret = xdp_push(task, src_handle);
-
 	} else if (src_handle->type == UV_UDP && ENABLE_SENDMMSG) {
 		int fd;
 		ret = uv_fileno(src_handle, &fd);
-		assert(!ret);
-		if (ret == 0) {
+		if (ret == 0)
 			udp_queue_push(fd, &ctx->req, task);
-		}
+		else
+			kr_assert(false);
 	} else {
 		ret = qr_task_send(task, source_session, &ctx->source.addr.ip, ctx->req.answer);
 	}
@@ -1326,7 +1374,7 @@ static int qr_task_finalize(struct qr_task *task, int state)
 		while (!session_tasklist_is_empty(source_session)) {
 			struct qr_task *t = session_tasklist_del_first(source_session, false);
 			struct request_ctx *c = t->ctx;
-			assert(c->source.session == source_session);
+			kr_assert(c->source.session == source_session);
 			c->source.session = NULL;
 			/* Don't finalize them as there can be other tasks
 			 * waiting for answer to this particular task.
@@ -1367,14 +1415,8 @@ static int udp_task_step(struct qr_task *task,
 
 static int tcp_task_waiting_connection(struct session *session, struct qr_task *task)
 {
-	assert(session_flags(session)->outgoing);
-	if (session_flags(session)->closing) {
-		/* Something went wrong. Better answer with KR_STATE_FAIL.
-		 * TODO: normally should not happen,
-		 * consider possibility to transform this into
-		 * assert(!session_flags(session)->closing). */
+	if (kr_fails_assert(session_flags(session)->outgoing && !session_flags(session)->closing))
 		return kr_error(EINVAL);
-	}
 	/* Add task to the end of list of waiting tasks.
 	 * It will be notified in on_connect() or qr_task_on_send(). */
 	int ret = session_waitinglist_push(session, task);
@@ -1386,18 +1428,11 @@ static int tcp_task_waiting_connection(struct session *session, struct qr_task *
 
 static int tcp_task_existing_connection(struct session *session, struct qr_task *task)
 {
-	assert(session_flags(session)->outgoing);
+	if (kr_fails_assert(session_flags(session)->outgoing && !session_flags(session)->closing))
+		return kr_error(EINVAL);
 	struct request_ctx *ctx = task->ctx;
 	struct worker_ctx *worker = ctx->worker;
 
-	if (session_flags(session)->closing) {
-		/* Something went wrong. Better answer with KR_STATE_FAIL.
-		 * TODO: normally should not happen,
-		 * consider possibility to transform this into
-		 * assert(!session_flags(session)->closing). */
-		return kr_error(EINVAL);
-	}
-
 	/* If there are any unsent queries, send it first. */
 	int ret = send_waiting(session);
 	if (ret != 0) {
@@ -1456,7 +1491,11 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 		return kr_error(EINVAL);
 	}
 	struct session *session = client->data;
-	assert(session_flags(session)->has_tls == has_tls);
+	if (kr_fails_assert(session_flags(session)->has_tls == has_tls)) {
+		tls_client_ctx_free(tls_ctx);
+		free(conn);
+		return kr_error(EINVAL);
+	}
 	if (has_tls) {
 		tls_client_ctx_set_session(tls_ctx, session);
 		session_tls_set_client_ctx(session, tls_ctx);
@@ -1487,7 +1526,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	}
 
 	struct kr_query *qry = task_get_last_pending_query(task);
-	WITH_VERBOSE (qry) {
+	if (kr_log_is_debug_qry(WORKER, qry)) {
 		const char *peer_str = kr_straddr(peer);
 		VERBOSE_MSG(qry, "=> connecting to: '%s'\n", peer_str ? peer_str : "");
 	}
@@ -1520,7 +1559,10 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 static int tcp_task_step(struct qr_task *task,
 			 const struct sockaddr *packet_source, knot_pkt_t *packet)
 {
-	assert(task->pending_count == 0);
+	if (kr_fails_assert(task->pending_count == 0)) {
+		subreq_finalize(task, packet_source, packet);
+		return qr_task_finalize(task, KR_STATE_FAIL);
+	}
 
 	/* target */
 	const struct sockaddr *addr = &task->transport->address.ip;
@@ -1579,7 +1621,8 @@ static int qr_task_step(struct qr_task *task,
 
 	/* Consume input and produce next query */
 	struct request_ctx *ctx = task->ctx;
-	assert(ctx);
+	if (kr_fails_assert(ctx))
+		return qr_task_finalize(task, KR_STATE_FAIL);
 	struct kr_request *req = &ctx->req;
 	struct worker_ctx *worker = ctx->worker;
 
@@ -1612,7 +1655,6 @@ static int qr_task_step(struct qr_task *task,
 		if (unlikely(++task->iter_count > KR_ITER_LIMIT ||
 			     task->timeouts >= KR_TIMEOUT_LIMIT)) {
 
-			#ifndef NOVERBOSELOG
 			struct kr_rplan *rplan = &req->rplan;
 			struct kr_query *last = kr_rplan_last(rplan);
 			if (task->iter_count > KR_ITER_LIMIT) {
@@ -1621,7 +1663,6 @@ static int qr_task_step(struct qr_task *task,
 			if (task->timeouts >= KR_TIMEOUT_LIMIT) {
 				VERBOSE_MSG(last, "canceling query due to exceeded timeout retries limit of %d\n", KR_TIMEOUT_LIMIT);
 			}
-			#endif
 
 			return qr_task_finalize(task, KR_STATE_FAIL);
 		}
@@ -1642,7 +1683,7 @@ static int qr_task_step(struct qr_task *task,
 	case KR_TRANSPORT_TLS:
 		return tcp_task_step(task, packet_source, packet);
 	default:
-		assert(0);
+		kr_assert(!EINVAL);
 		return kr_error(EINVAL);
 	}
 }
@@ -1697,8 +1738,11 @@ int worker_submit(struct session *session,
 	    (is_query == is_outgoing)) {
 		if (!is_outgoing) {
 			the_worker->stats.dropped += 1;
-			if (http_ctx)
+			if (http_ctx) {
+				struct http_stream stream = queue_head(http_ctx->streams);
+				http_free_headers(stream.headers);
 				queue_pop(http_ctx->streams);
+			}
 		}
 		return kr_error(EILSEQ);
 	}
@@ -1739,12 +1783,14 @@ int worker_submit(struct session *session,
 					(int)id);
 			return kr_error(ENOENT);
 		}
-		assert(!session_flags(session)->closing);
+		if (kr_fails_assert(!session_flags(session)->closing))
+			return kr_error(EINVAL);
 		addr = peer;
 		/* Note recieve time for RTT calculation */
 		task->recv_time = kr_now();
 	}
-	assert(uv_is_closing(session_get_handle(session)) == false);
+	if (kr_fails_assert(!uv_is_closing(session_get_handle(session))))
+		return kr_error(EINVAL);
 
 	/* Packet was successfully parsed.
 	 * Task was created (found). */
@@ -1757,19 +1803,22 @@ int worker_submit(struct session *session,
 static int map_add_tcp_session(map_t *map, const struct sockaddr* addr,
 			       struct session *session)
 {
-	assert(map && addr);
+	if (kr_fails_assert(map && addr))
+		return kr_error(EINVAL);
 	const char *key = tcpsess_key(addr);
-	assert(key);
-	assert(map_contains(map, key) == 0);
+	if (kr_fails_assert(key && map_contains(map, key) == 0))
+		return kr_error(EINVAL);
 	int ret = map_set(map, key, session);
 	return ret ? kr_error(EINVAL) : kr_ok();
 }
 
 static int map_del_tcp_session(map_t *map, const struct sockaddr* addr)
 {
-	assert(map && addr);
+	if (kr_fails_assert(map && addr))
+		return kr_error(EINVAL);
 	const char *key = tcpsess_key(addr);
-	assert(key);
+	if (kr_fails_assert(key))
+		return kr_error(EINVAL);
 	int ret = map_del(map, key);
 	return ret ? kr_error(ENOENT) : kr_ok();
 }
@@ -1777,9 +1826,11 @@ static int map_del_tcp_session(map_t *map, const struct sockaddr* addr)
 static struct session* map_find_tcp_session(map_t *map,
 					    const struct sockaddr *addr)
 {
-	assert(map && addr);
+	if (kr_fails_assert(map && addr))
+		return NULL;
 	const char *key = tcpsess_key(addr);
-	assert(key);
+	if (kr_fails_assert(key))
+		return NULL;
 	struct session* ret = map_get(map, key);
 	return ret;
 }
@@ -1788,19 +1839,12 @@ int worker_add_tcp_connected(struct worker_ctx *worker,
 				    const struct sockaddr* addr,
 				    struct session *session)
 {
-#ifndef NDEBUG
-	assert(addr);
-	const char *key = tcpsess_key(addr);
-	assert(key);
-	assert(map_contains(&worker->tcp_connected, key) == 0);
-#endif
 	return map_add_tcp_session(&worker->tcp_connected, addr, session);
 }
 
 int worker_del_tcp_connected(struct worker_ctx *worker,
 				    const struct sockaddr* addr)
 {
-	assert(addr && tcpsess_key(addr));
 	return map_del_tcp_session(&worker->tcp_connected, addr);
 }
 
@@ -1814,19 +1858,12 @@ static int worker_add_tcp_waiting(struct worker_ctx *worker,
 				  const struct sockaddr* addr,
 				  struct session *session)
 {
-#ifndef NDEBUG
-	assert(addr);
-	const char *key = tcpsess_key(addr);
-	assert(key);
-	assert(map_contains(&worker->tcp_waiting, key) == 0);
-#endif
 	return map_add_tcp_session(&worker->tcp_waiting, addr, session);
 }
 
 int worker_del_tcp_waiting(struct worker_ctx *worker,
 			   const struct sockaddr* addr)
 {
-	assert(addr && tcpsess_key(addr));
 	return map_del_tcp_session(&worker->tcp_waiting, addr);
 }
 
@@ -1838,9 +1875,8 @@ struct session* worker_find_tcp_waiting(struct worker_ctx *worker,
 
 int worker_end_tcp(struct session *session)
 {
-	if (!session) {
+	if (!session)
 		return kr_error(EINVAL);
-	}
 
 	session_timer_stop(session);
 
@@ -1864,7 +1900,7 @@ int worker_end_tcp(struct session *session)
 
 	while (!session_waitinglist_is_empty(session)) {
 		struct qr_task *task = session_waitinglist_pop(session, false);
-		assert(task->refs > 1);
+		kr_assert(task->refs > 1);
 		session_tasklist_del(session, task);
 		if (session_flags(session)->outgoing) {
 			if (task->ctx->req.options.FORWARD) {
@@ -1879,7 +1915,7 @@ int worker_end_tcp(struct session *session)
 			}
 			qr_task_step(task, NULL, NULL);
 		} else {
-			assert(task->ctx->source.session == session);
+			kr_assert(task->ctx->source.session == session);
 			task->ctx->source.session = NULL;
 		}
 		worker_task_unref(task);
@@ -1895,7 +1931,7 @@ int worker_end_tcp(struct session *session)
 			}
 			qr_task_step(task, NULL, NULL);
 		} else {
-			assert(task->ctx->source.session == session);
+			kr_assert(task->ctx->source.session == session);
 			task->ctx->source.session = NULL;
 		}
 		worker_task_unref(task);
@@ -1953,17 +1989,14 @@ knot_pkt_t *worker_resolve_mk_pkt(const char *qname_str, uint16_t qtype, uint16_
 struct qr_task *worker_resolve_start(knot_pkt_t *query, struct kr_qflags options)
 {
 	struct worker_ctx *worker = the_worker;
-	if (!worker || !query) {
-		assert(!EINVAL);
+	if (kr_fails_assert(worker && query))
 		return NULL;
-	}
 
 
 	struct request_ctx *ctx = request_create(worker, NULL, NULL, NULL, NULL, NULL,
 						 worker->next_request_uid);
-	if (!ctx) {
+	if (!ctx)
 		return NULL;
-	}
 
 	/* Create task */
 	struct qr_task *task = qr_task_create(ctx);
@@ -1984,9 +2017,8 @@ struct qr_task *worker_resolve_start(knot_pkt_t *query, struct kr_qflags options
 	}
 
 	worker->next_request_uid += 1;
-	if (worker->next_request_uid == 0) {
+	if (worker->next_request_uid == 0)
 		worker->next_request_uid = UINT16_MAX + 1;
-	}
 
 	/* Set options late, as qr_task_start() -> kr_resolve_begin() rewrite it. */
 	kr_qflags_set(&task->ctx->req.options, options);
@@ -1995,9 +2027,8 @@ struct qr_task *worker_resolve_start(knot_pkt_t *query, struct kr_qflags options
 
 int worker_resolve_exec(struct qr_task *task, knot_pkt_t *query)
 {
-	if (!task) {
+	if (!task)
 		return kr_error(EINVAL);
-	}
 	return qr_task_step(task, NULL, query);
 }
 
@@ -2008,9 +2039,8 @@ int worker_task_numrefs(const struct qr_task *task)
 
 struct kr_request *worker_task_request(struct qr_task *task)
 {
-	if (!task || !task->ctx) {
+	if (!task || !task->ctx)
 		return NULL;
-	}
 
 	return &task->ctx->req;
 }
@@ -2123,7 +2153,8 @@ static inline void reclaim_mp_freelist(mp_freelist_t *list)
 void worker_deinit(void)
 {
 	struct worker_ctx *worker = the_worker;
-	assert(worker);
+	if (kr_fails_assert(worker))
+		return;
 	if (worker->z_import != NULL) {
 		zi_free(worker->z_import);
 		worker->z_import = NULL;
@@ -2133,6 +2164,10 @@ void worker_deinit(void)
 	trie_free(worker->subreq_out);
 	worker->subreq_out = NULL;
 
+	for (int i = 0; i < worker->doh_qry_headers.len; i++)
+		free((void *)worker->doh_qry_headers.at[i]);
+	array_clear(worker->doh_qry_headers);
+
 	reclaim_mp_freelist(&worker->pool_mp);
 	mp_delete(worker->pkt_pool.ctx);
 	worker->pkt_pool.ctx = NULL;
@@ -2142,8 +2177,8 @@ void worker_deinit(void)
 
 int worker_init(struct engine *engine, int worker_count)
 {
-	assert(engine && engine->L);
-	assert(the_worker == NULL);
+	if (kr_fails_assert(engine && engine->L && the_worker == NULL))
+		return kr_error(EINVAL);
 	kr_bindings_register(engine->L);
 
 	/* Create main worker. */
@@ -2167,6 +2202,8 @@ int worker_init(struct engine *engine, int worker_count)
 	worker->out_addr4.sin_family = AF_UNSPEC;
 	worker->out_addr6.sin6_family = AF_UNSPEC;
 
+	array_init(worker->doh_qry_headers);
+
 	int ret = worker_reserve(worker, MP_FREELIST_SIZE);
 	if (ret) return ret;
 	worker->next_request_uid = UINT16_MAX + 1;
@@ -2181,7 +2218,7 @@ int worker_init(struct engine *engine, int worker_count)
 		lua_pushstring(engine->L, inst_name);
 	} else {
 		ret = asprintf(&pid_str, "%ld", (long)pid);
-		assert(ret > 0);
+		kr_assert(ret > 0);
 		lua_pushstring(engine->L, pid_str);
 	}
 	lua_setfield(engine->L, -2, "id");
