@@ -47,7 +47,7 @@
 #include "lib/generic/map.h"
 #include "lib/generic/array.h"
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "zimport", __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, ZIMPORT, __VA_ARGS__)
 
 /* Pause between parse and import stages, milliseconds.
  * See comment in zi_zone_import() */
@@ -236,8 +236,8 @@ static int zi_rrset_find_put(struct zone_import_ctx *z_import,
 static int zi_rrset_put(struct zone_import_ctx *z_import, knot_pkt_t *pkt,
 			knot_rrset_t *rr)
 {
-	assert(rr);
-	assert(rr->type != KNOT_RRTYPE_RRSIG);
+	if (kr_fails_assert(rr && rr->type != KNOT_RRTYPE_RRSIG))
+		return -1;
 	int err = knot_pkt_put(pkt, 0, rr, 0);
 	if (err != KNOT_EOK) {
 		return -1;
@@ -450,26 +450,29 @@ static void zi_zone_process(uv_timer_t* handle)
 {
 	zone_import_ctx_t *z_import = (zone_import_ctx_t *)handle->data;
 
-	assert(z_import->worker);
-
 	size_t failed = 0;
 	size_t ns_imported = 0;
 	size_t other_imported = 0;
 
+	if (kr_fails_assert(z_import->worker)) {
+		failed = 1;
+		goto finish;
+	}
+
 	/* At the moment import of root zone only is supported.
 	 * Check the name of the parsed zone.
 	 * TODO - implement importing of arbitrary zone. */
 	KR_DNAME_GET_STR(zone_name_str, z_import->origin);
 
 	if (strcmp(".", zone_name_str) != 0) {
-		kr_log_error("[zimport] unexpected zone name `%s` (root zone expected), fail\n",
+		kr_log_error(ZIMPORT, "unexpected zone name `%s` (root zone expected), fail\n",
 			     zone_name_str);
 		failed = 1;
 		goto finish;
 	}
 
 	if (z_import->rrset_sorted.len <= 0) {
-		kr_log_error("[zimport] zone `%s` is empty\n", zone_name_str);
+		kr_log_error(ZIMPORT, "zone `%s` is empty\n", zone_name_str);
 		goto finish;
 	}
 
@@ -487,7 +490,7 @@ static void zi_zone_process(uv_timer_t* handle)
 	knot_rrset_t *rr_key = map_get(&z_import->rrset_indexed, key);
 	if (!rr_key) {
 		/* DNSKEY MUST be here. If not found - fail. */
-		kr_log_error("[zimport] DNSKEY not found for `%s`, fail\n", zone_name_str);
+		kr_log_error(ZIMPORT, "DNSKEY not found for `%s`, fail\n", zone_name_str);
 		failed = 1;
 		goto finish;
 	}
@@ -496,7 +499,7 @@ static void zi_zone_process(uv_timer_t* handle)
 	map_t *trust_anchors = &z_import->worker->engine->resolver.trust_anchors;
 	knot_rrset_t *rr_ta = kr_ta_get(trust_anchors, z_import->origin);
 	if (!rr_ta) {
-		kr_log_error("[zimport] error: TA for zone `%s` vanished, fail", zone_name_str);
+		kr_log_error(ZIMPORT, "error: TA for zone `%s` vanished, fail", zone_name_str);
 		failed = 1;
 		goto finish;
 	}
@@ -516,7 +519,7 @@ static void zi_zone_process(uv_timer_t* handle)
 
 	int res = zi_rrset_import(z_import, rr_key);
 	if (res != 0) {
-		kr_log_error("[zimport] import failed: qname: '%s' type: '%s'\n",
+		kr_log_error(ZIMPORT, "import failed: qname: '%s' type: '%s'\n",
 			    kname_str, ktype_str);
 		failed = 1;
 		goto finish;
@@ -593,7 +596,7 @@ finish:
 	if (failed != 0) {
 		if (ns_imported == 0 && other_imported == 0) {
 			import_state = -1;
-			kr_log_error("[zimport] import failed; zone `%s` \n", zone_name_str);
+			kr_log_error(ZIMPORT, "import failed; zone `%s` \n", zone_name_str);
 		} else {
 			import_state = 1;
 		}
@@ -612,13 +615,13 @@ static int zi_record_store(zs_scanner_t *s)
 {
 	if (s->r_data_length > UINT16_MAX) {
 		/* Due to knot_rrset_add_rdata(..., const uint16_t size, ...); */
-		kr_log_error("[zscanner] line %"PRIu64": rdata is too long\n",
+		kr_log_error(ZSCANNER, "line %"PRIu64": rdata is too long\n",
 				s->line_counter);
 		return -1;
 	}
 
 	if (knot_dname_size(s->r_owner) != strlen((const char *)(s->r_owner)) + 1) {
-		kr_log_error("[zscanner] line %"PRIu64
+		kr_log_error(ZSCANNER, "line %"PRIu64
 				": owner name contains zero byte, skip\n",
 				s->line_counter);
 		return 0;
@@ -629,14 +632,14 @@ static int zi_record_store(zs_scanner_t *s)
 	knot_rrset_t *new_rr = knot_rrset_new(s->r_owner, s->r_type, s->r_class,
 					      s->r_ttl, &z_import->pool);
 	if (!new_rr) {
-		kr_log_error("[zscanner] line %"PRIu64": error creating rrset\n",
+		kr_log_error(ZSCANNER, "line %"PRIu64": error creating rrset\n",
 				s->line_counter);
 		return -1;
 	}
 	int res = knot_rrset_add_rdata(new_rr, s->r_data, s->r_data_length,
 				       &z_import->pool);
 	if (res != KNOT_EOK) {
-		kr_log_error("[zscanner] line %"PRIu64": error adding rdata to rrset\n",
+		kr_log_error(ZSCANNER, "line %"PRIu64": error adding rdata to rrset\n",
 				s->line_counter);
 		return -1;
 	}
@@ -650,7 +653,7 @@ static int zi_record_store(zs_scanner_t *s)
 	res = kr_rrkey(key, new_rr->rclass, new_rr->owner, new_rr->type,
 		       additional_key_field);
 	if (res <= 0) {
-		kr_log_error("[zscanner] line %"PRIu64": error constructing rrkey\n",
+		kr_log_error(ZSCANNER, "line %"PRIu64": error constructing rrkey\n",
 				s->line_counter);
 		return -1;
 	}
@@ -663,7 +666,7 @@ static int zi_record_store(zs_scanner_t *s)
 		res = map_set(&z_import->rrset_indexed, key, new_rr);
 	}
 	if (res != 0) {
-		kr_log_error("[zscanner] line %"PRIu64": error saving parsed rrset\n",
+		kr_log_error(ZSCANNER, "line %"PRIu64": error saving parsed rrset\n",
 				s->line_counter);
 		return -1;
 	}
@@ -689,29 +692,29 @@ static int zi_state_parsing(zs_scanner_t *s)
 			}
 			break;
 		case ZS_STATE_ERROR:
-			kr_log_error("[zscanner] line: %"PRIu64
+			kr_log_error(ZSCANNER, "line: %"PRIu64
 				     ": parse error; code: %i ('%s')\n",
 				     s->line_counter, s->error.code,
 				     zs_strerror(s->error.code));
 			return -1;
 		case ZS_STATE_INCLUDE:
-			kr_log_error("[zscanner] line: %"PRIu64
+			kr_log_error(ZSCANNER, "line: %"PRIu64
 				     ": INCLUDE is not supported\n",
 				     s->line_counter);
 			return -1;
 		case ZS_STATE_EOF:
 		case ZS_STATE_STOP:
 			if (empty) {
-				kr_log_error("[zimport] empty zone file\n");
+				kr_log_error(ZIMPORT, "empty zone file\n");
 				return -1;
 			}
 			if (!((zone_import_ctx_t *) s->process.data)->origin) {
-				kr_log_error("[zimport] zone file doesn't contain SOA record\n");
+				kr_log_error(ZIMPORT, "zone file doesn't contain SOA record\n");
 				return -1;
 			}
 			return (s->error.counter == 0) ? 0 : -1;
 		default:
-			kr_log_error("[zscanner] line: %"PRIu64
+			kr_log_error(ZSCANNER, "line: %"PRIu64
 				     ": unexpected parse state: %i\n",
 				     s->line_counter, s->state);
 			return -1;
@@ -725,13 +728,12 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 		   const char *zone_file, const char *origin,
 		   uint16_t rclass, uint32_t ttl)
 {
-	assert (z_import != NULL && "[zimport] empty <z_import> parameter");
-	assert (z_import->worker != NULL && "[zimport] invalid <z_import> parameter\n");
-	assert (zone_file != NULL && "[zimport] empty <zone_file> parameter\n");
+	if (kr_fails_assert(z_import && z_import->worker && zone_file))
+		return -1;
 
 	zs_scanner_t *s = malloc(sizeof(zs_scanner_t));
 	if (s == NULL) {
-		kr_log_error("[zscanner] error creating instance of zone scanner (malloc() fails)\n");
+		kr_log_error(ZSCANNER, "error creating instance of zone scanner (malloc() fails)\n");
 		return -1;
 	}
 
@@ -739,7 +741,7 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 	 * so don't print error code as it meaningless. */
 	int res = zs_init(s, origin, rclass, ttl);
 	if (res != 0) {
-		kr_log_error("[zscanner] error initializing zone scanner instance, error: %i (%s)\n",
+		kr_log_error(ZSCANNER, "error initializing zone scanner instance, error: %i (%s)\n",
 			     s->error.code, zs_strerror(s->error.code));
 		free(s);
 		return -1;
@@ -747,7 +749,7 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 
 	res = zs_set_input_file(s, zone_file);
 	if (res != 0) {
-		kr_log_error("[zscanner] error opening zone file `%s`, error: %i (%s)\n",
+		kr_log_error(ZSCANNER, "error opening zone file `%s`, error: %i (%s)\n",
 			     zone_file, s->error.code, zs_strerror(s->error.code));
 		zs_deinit(s);
 		free(s);
@@ -758,7 +760,7 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 	 * Parsing as well error processing will be performed in zi_state_parsing().
 	 * Store pointer to zone import context for further use. */
 	if (zs_set_processing(s, NULL, NULL, (void *)z_import) != 0) {
-		kr_log_error("[zscanner] zs_set_processing() failed for zone file `%s`, "
+		kr_log_error(ZSCANNER, "zs_set_processing() failed for zone file `%s`, "
 				"error: %i (%s)\n",
 				zone_file, s->error.code, zs_strerror(s->error.code));
 		zs_deinit(s);
@@ -782,7 +784,7 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 				/* For now - fail.
 				 * TODO - query DS and continue after answer had been obtained. */
 				KR_DNAME_GET_STR(zone_name_str, z_import->origin);
-				kr_log_error("[zimport] no TA found for `%s`, fail\n", zone_name_str);
+				kr_log_error(ZIMPORT, "no TA found for `%s`, fail\n", zone_name_str);
 				ret = 1;
 			}
 			elapsed = kr_now() - z_import->start_timestamp;
@@ -793,7 +795,7 @@ int zi_zone_import(struct zone_import_ctx *z_import,
 	free(s);
 
 	if (ret != 0) {
-		kr_log_error("[zscanner] error parsing zone file `%s`\n", zone_file);
+		kr_log_error(ZSCANNER, "error parsing zone file `%s`\n", zone_file);
 		z_import->started = false;
 		return ret;
 	}
