@@ -14,7 +14,7 @@
 
 #include "lib/utils.h"
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), "slct", __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), SELECTION, __VA_ARGS__)
 
 #define DEFAULT_TIMEOUT 400
 #define MAX_TIMEOUT 10000
@@ -55,7 +55,7 @@ static const char *kr_selection_error_str(enum kr_selection_error err) {
 		case KR_SELECTION_NUMBER_OF_ERRORS: break; // not a valid code
 	#undef X
 	}
-	assert(false); // we want to define all; compiler helps by -Wswitch (no default:)
+	kr_assert(false); // we want to define all; compiler helps by -Wswitch (no default:)
 	return NULL;
 }
 
@@ -152,8 +152,8 @@ struct rtt_state get_rtt_state(const uint8_t *ip, size_t len,
 
 	if (cache->api->read(db, stats, &key, &value, 1)) {
 		state = default_rtt_state;
-	} else if (value.len != sizeof(struct rtt_state)) {
-		assert(false); // shouldn't happen but let's be more robust
+	} else if (kr_fails_assert(value.len == sizeof(struct rtt_state))) {
+		// shouldn't happen but let's be more robust
 		state = default_rtt_state;
 	} else { // memcpy is safe for unaligned case (on non-x86)
 		memcpy(&state, value.data, sizeof(state));
@@ -195,7 +195,7 @@ void bytes_to_ip(uint8_t *bytes, size_t len, uint16_t port, union inaddr *dst)
 		dst->ip6.sin6_port = htons(port);
 		break;
 	default:
-		assert(0);
+		kr_assert(false);
 	}
 }
 
@@ -207,7 +207,7 @@ uint8_t *ip_to_bytes(const union inaddr *src, size_t len)
 	case sizeof(struct in6_addr):
 		return (uint8_t *)&src->ip6.sin6_addr;
 	default:
-		assert(0);
+		kr_assert(false);
 		return NULL;
 	}
 }
@@ -263,7 +263,7 @@ static void invalidate_dead_upstream(struct address_state *state,
 			// period when we don't want to use the address
 			state->generation = -1;
 		} else {
-			assert(now >= rs->dead_since + retry_timeout);
+			kr_assert(now >= rs->dead_since + retry_timeout);
 			// we allow to retry the server now
 			// TODO: perhaps tweak *rs?
 		}
@@ -327,8 +327,7 @@ void update_address_state(struct address_state *state, union inaddr *address,
 		state, qry->request->ctx->cache_rtt_tout_retry_interval);
 #ifdef SELECTION_CHOICE_LOGGING
 	// This is sometimes useful for debugging, but usually too verbose
-	WITH_VERBOSE(qry)
-	{
+	if (kr_log_is_debug_qry(SELECTION, qry)) {
 		const char *ns_str = kr_straddr(&address->ip);
 		VERBOSE_MSG(qry, "rtt of %s is %d, variance is %d\n", ns_str,
 			    state->rtt_state.srtt, state->rtt_state.variance);
@@ -486,7 +485,7 @@ struct kr_transport *select_transport(struct choice choices[], int choices_len,
 			port = KR_DNS_PORT;
 			break;
 		default:
-			assert(0);
+			kr_assert(false);
 			return NULL;
 		}
 	}
@@ -502,7 +501,7 @@ struct kr_transport *select_transport(struct choice choices[], int choices_len,
 		transport->address.ip6.sin6_port = htons(port);
 		break;
 	default:
-		assert(0);
+		kr_assert(false);
 		return NULL;
 	}
 
@@ -538,18 +537,17 @@ void update_rtt(struct kr_query *qry, struct address_state *addr_state,
 	if (transport->address_len == sizeof(struct in6_addr))
 		no6_success(qry);
 
-	WITH_VERBOSE(qry)
-	{
-	KR_DNAME_GET_STR(ns_name, transport->ns_name);
-	KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
-	const char *ns_str = kr_straddr(&transport->address.ip);
+	if (kr_log_is_debug_qry(SELECTION, qry)) {
+		KR_DNAME_GET_STR(ns_name, transport->ns_name);
+		KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
+		const char *ns_str = kr_straddr(&transport->address.ip);
 
-	VERBOSE_MSG(
-		qry,
-		"=> id: '%05u' updating: '%s'@'%s' zone cut: '%s'"
-		" with rtt %u to srtt: %d and variance: %d \n",
-		qry->id, ns_name, ns_str ? ns_str : "", zonecut_str,
-		rtt, new_rtt_state.srtt, new_rtt_state.variance);
+		VERBOSE_MSG(
+			qry,
+			"=> id: '%05u' updating: '%s'@'%s' zone cut: '%s'"
+			" with rtt %u to srtt: %d and variance: %d \n",
+			qry->id, ns_name, ns_str ? ns_str : "", zonecut_str,
+			rtt, new_rtt_state.srtt, new_rtt_state.variance);
 	}
 }
 
@@ -671,25 +669,25 @@ void error(struct kr_query *qry, struct address_state *addr_state,
 		addr_state->broken = true;
 		break;
 	default:
-		assert(0);
-		break;
+		kr_assert(false);
+		return;
 	}
 
 	addr_state->error_count++;
 	addr_state->errors[sel_error]++;
 
-	WITH_VERBOSE(qry)
-	{
-	KR_DNAME_GET_STR(ns_name, transport->ns_name);
-	KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
-	const char *ns_str = kr_straddr(&transport->address.ip);
-	const char *err_str = kr_selection_error_str(sel_error);
+	if (kr_log_is_debug_qry(SELECTION, qry)) {
+		KR_DNAME_GET_STR(ns_name, transport->ns_name);
+		KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
+		const char *ns_str = kr_straddr(&transport->address.ip);
+		const char *err_str = kr_selection_error_str(sel_error);
 
-	VERBOSE_MSG(
-		qry,
-		"=> id: '%05u' noting selection error: '%s'@'%s' zone cut: '%s' error: %d %s\n",
-		qry->id, ns_name, ns_str ? ns_str : "", zonecut_str,
-		sel_error, err_str ? err_str : "??");
+		VERBOSE_MSG(
+			qry,
+			"=> id: '%05u' noting selection error: '%s'@'%s'"
+			" zone cut: '%s' error: %d %s\n",
+			qry->id, ns_name, ns_str ? ns_str : "",
+			zonecut_str, sel_error, err_str ? err_str : "??");
 	}
 }
 
