@@ -77,18 +77,13 @@ function policy.MIRROR(target)
 end
 
 -- Override the list of nameservers (forwarders)
-local function set_nslist(qry, list)
+local function set_nslist(req, list)
 	local ns_i = 0
 	for _, ns in ipairs(list) do
-		-- kr_nsrep_set() can return kr_error(ENOENT), it's OK
-		if ffi.C.kr_nsrep_set(qry, ns_i, ns) == 0 then
+		if ffi.C.kr_forward_add_target(req, ns) == 0 then
 			ns_i = ns_i + 1
 		end
 	end
-	-- If less than maximum NSs, insert guard to terminate the list
-	if ns_i < 3 then
-		assert(ffi.C.kr_nsrep_set(qry, ns_i, nil) == 0);
-	end
 	if ns_i == 0 then
 		-- would use assert() but don't want to compose the message if not triggered
 		error('no usable address in NS set (check net.ipv4 and '
@@ -102,7 +97,6 @@ function policy.STUB(target)
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
 			table.insert(list, addr2sock(v, 53))
-			assert(#list <= 4, 'at most 4 STUB targets are supported')
 		end
 	else
 		table.insert(list, addr2sock(target, 53))
@@ -112,7 +106,7 @@ function policy.STUB(target)
 		-- Switch mode to stub resolver, do not track origin zone cut since it's not real authority NS
 		qry.flags.STUB = true
 		qry.flags.ALWAYS_CUT = false
-		set_nslist(qry, list)
+		set_nslist(req, list)
 		return state
 	end
 end
@@ -123,7 +117,6 @@ function policy.FORWARD(target)
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
 			table.insert(list, addr2sock(v, 53))
-			assert(#list <= 4, 'at most 4 FORWARD targets are supported')
 		end
 	else
 		table.insert(list, addr2sock(target, 53))
@@ -136,7 +129,7 @@ function policy.FORWARD(target)
 		qry.flags.ALWAYS_CUT = false
 		qry.flags.NO_MINIMIZE = true
 		qry.flags.AWAIT_CUT = true
-		set_nslist(qry, list)
+		set_nslist(req, list)
 		return state
 	end
 end
@@ -145,8 +138,6 @@ end
 function policy.TLS_FORWARD(targets)
 	if type(targets) ~= 'table' or #targets < 1 then
 		error('TLS_FORWARD argument must be a non-empty table')
-	elseif #targets > 4 then
-		error('TLS_FORWARD supports at most four targets (in a single call)')
 	end
 
 	local sockaddr_c_set = {}
@@ -182,7 +173,7 @@ function policy.TLS_FORWARD(targets)
 		qry.flags.AWAIT_CUT = true
 		req.options.TCP = true
 		qry.flags.TCP = true
-		set_nslist(qry, nslist)
+		set_nslist(req, nslist)
 		return state
 	end
 end
@@ -202,53 +193,71 @@ end
 -- Set and clear some query flags
 function policy.FLAGS(opts_set, opts_clear)
 	return function(_, req)
+		-- We assume to be running in the begin phase, so to truly apply
+		-- to the whole request we need to change both kr_request and kr_query.
 		local qry = req:current()
-		ffi.C.kr_qflags_set  (qry.flags, kres.mk_qflags(opts_set   or {}))
-		ffi.C.kr_qflags_clear(qry.flags, kres.mk_qflags(opts_clear or {}))
+		for _, flags in pairs({qry.flags, req.options}) do
+			ffi.C.kr_qflags_set  (flags, kres.mk_qflags(opts_set   or {}))
+			ffi.C.kr_qflags_clear(flags, kres.mk_qflags(opts_clear or {}))
+		end
 		return nil -- chain rule
 	end
 end
 
+local function mkauth_soa(answer, dname, mname, ttl)
+	if mname == nil then
+		mname = dname
+	end
+	return answer:put(dname, ttl or 10800, answer:qclass(), kres.type.SOA,
+		mname .. '\6nobody\7invalid\0\0\0\0\1\0\0\14\16\0\0\4\176\0\9\58\128\0\0\42\48')
+end
+
 -- Create answer with passed arguments
 function policy.ANSWER(rtable, nodata)
 	return function(_, req)
 		local qry = req:current()
-		local answer = req.answer
 		local data = rtable[qry.stype]
+		if data == nil and nodata ~= true then
+			return nil
+		end
+		-- now we're certain we want to generate an answer
 
+		local answer = req:ensure_answer()
+		if answer == nil then return nil end
 		ffi.C.kr_pkt_make_auth_header(answer)
+		local ttl = (data or {}).ttl or 1
+		answer:rcode(kres.rcode.NOERROR)
 
-		if data == nil then
-			if nodata == true then
-				answer:rcode(kres.rcode.NOERROR)
-				return kres.DONE
+		if data == nil then -- want NODATA, i.e. just a SOA
+			answer:begin(kres.section.AUTHORITY)
+			local soa = rtable[kres.type.SOA]
+			if soa ~= nil then
+				answer:put(qry.sname, soa.ttl or ttl, qry.sclass, kres.type.SOA,
+							soa.rdata[1] or soa.rdata)
+			else
+				mkauth_soa(answer, kres.dname2wire(qry.sname), nil, ttl)
 			end
 		else
-			local ttl = data.ttl or 1
-
-			answer:rcode(kres.rcode.NOERROR)
 			answer:begin(kres.section.ANSWER)
-			answer:put(qry.sname, ttl, qry.sclass, qry.stype, data.rdata)
-
-			return kres.DONE
+			if type(data.rdata) == 'table' then
+				for _, rdato in ipairs(data.rdata) do
+					answer:put(qry.sname, ttl, qry.sclass, qry.stype, rdato)
+				end
+			else
+				answer:put(qry.sname, ttl, qry.sclass, qry.stype, data.rdata)
+			end
 		end
+		return kres.DONE
 	end
 end
 
-local function mkauth_soa(answer, dname, mname)
-	if mname == nil then
-		mname = dname
-	end
-	return answer:put(dname, 10800, answer:qclass(), kres.type.SOA,
-		mname .. '\6nobody\7invalid\0\0\0\0\1\0\0\14\16\0\0\4\176\0\9\58\128\0\0\42\48')
-end
-
 local dname_localhost = todname('localhost.')
 
 -- Rule for localhost. zone; see RFC6303, sec. 3
 local function localhost(_, req)
 	local qry = req:current()
-	local answer = req.answer
+	local answer = req:ensure_answer()
+	if answer == nil then return nil end
 	ffi.C.kr_pkt_make_auth_header(answer)
 
 	local is_exact = ffi.C.knot_dname_is_equal(qry.sname, dname_localhost)
@@ -280,7 +289,8 @@ local dname_rev4_localhost_apex = todname('127.in-addr.arpa');
 -- TODO: much of this would better be left to the hints module (or coordinated).
 local function localhost_reversed(_, req)
 	local qry = req:current()
-	local answer = req.answer
+	local answer = req:ensure_answer()
+	if answer == nil then return nil end
 
 	-- classify qry.sname:
 	local is_exact   -- exact dname for localhost
@@ -378,7 +388,6 @@ end
 local function rpz_parse(action, path)
 	local rules = {}
 	local new_actions = {}
-	local origin = '\0'
 	local action_map = {
 		-- RPZ Policy Actions
 		['\0'] = action,
@@ -388,22 +397,17 @@ local function rpz_parse(action, path)
 		['\012rpz-tcp-only\0'] = policy.TC,
 		-- Policy triggers @NYI@
 	}
-	local unsupp_rrs = function (rtype)
-		local set = {
-			kres.type.DNAME,
-			kres.type.NS,
-			kres.type.SOA,
-			kres.type.DNSKEY,
-			kres.type.DS,
-			kres.type.RRSIG,
-			kres.type.NSEC,
-			kres.type.NSEC3,
-		}
-		for _, l in pairs(set) do
-			if rtype == l then return true end
-		end
-		return false
-	end
+	-- RR types to be skipped; boolean denoting whether to throw a warning even for RPZ apex.
+	local rrtype_bad = {
+		[kres.type.DNAME]  = true,
+		[kres.type.NS]     = false,
+		[kres.type.SOA]    = false,
+		[kres.type.DNSKEY] = true,
+		[kres.type.DS]     = true,
+		[kres.type.RRSIG]  = true,
+		[kres.type.NSEC]   = true,
+		[kres.type.NSEC3]  = true,
+	}
 	local parser = require('zonefile').new()
 	local ok, errstr = parser:open(path)
 	if not ok then
@@ -412,7 +416,7 @@ local function rpz_parse(action, path)
 	while true do
 		ok, errstr = parser:parse()
 		if errstr then
-			log('[poli] RPZ %s:%d: %s', path, tonumber(parser.line_counter), errstr)
+			log_info(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d: %s', path, tonumber(parser.line_counter), errstr)
 		end
 		if not ok then break end
 
@@ -420,39 +424,47 @@ local function rpz_parse(action, path)
 		local rdata = ffi.string(parser.r_data, parser.r_data_length)
 		ffi.C.knot_dname_to_lower(full_name)
 
-		if (parser.r_type == kres.type.SOA) then
-			origin = ffi.gc(ffi.C.knot_dname_copy(full_name, nil), ffi.C.free)
+		local prefix_labels = ffi.C.knot_dname_in_bailiwick(full_name, parser.zone_origin)
+		if prefix_labels < 0 then
+			log_info(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d: RR owner "%s" outside the zone (ignored)',
+				path, tonumber(parser.line_counter), kres.dname2str(full_name))
 			goto continue
 		end
 
-		local prefix_labels = ffi.C.knot_dname_in_bailiwick(full_name, origin)
-		local name
-		if prefix_labels > 0 then
-			local bytes = 0
-			for _=1,prefix_labels do
-				bytes = bytes + 1 + full_name[bytes]
-			end
-			name = ffi.string(full_name, bytes)
-			name = name..'\0'
-		else
-			name = ffi.string(full_name, parser.r_owner_length)
-		end
+		local bytes = ffi.C.knot_dname_size(full_name) - ffi.C.knot_dname_size(parser.zone_origin)
+		local name = ffi.string(full_name, bytes) .. '\0'
 
 		if parser.r_type == kres.type.CNAME then
 			if action_map[rdata] then
 				rules[name] = action_map[rdata]
 			else
-				log('[poli] RPZ %s:%d: CNAME with custom target in RPZ is not supported', path, tonumber(parser.line_counter))
+				log_info(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d: CNAME with custom target in RPZ is not supported yet (ignored)',
+					path, tonumber(parser.line_counter))
 			end
 		else
-			-- Warn when NYI
 			if #name then
-				if unsupp_rrs(parser.r_type) then
-					log('[poli] RPZ %s:%d: RR type %s is not allowed in RPZ', path, tonumber(parser.line_counter),
-					    kres.tostring.type[parser.r_type])
-				else
+				local is_bad = rrtype_bad[parser.r_type]
+				if is_bad == true or (is_bad == false and prefix_labels ~= 0) then
+					log_warn(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d warning: RR type %s is not allowed in RPZ (ignored)',
+						path, tonumber(parser.line_counter), kres.tostring.type[parser.r_type])
+				elseif is_bad == nil then
 					if new_actions[name] == nil then new_actions[name] = {} end
-					new_actions[name][parser.r_type] = { ttl=parser.r_ttl, rdata=rdata }
+					local act = new_actions[name][parser.r_type]
+					if act == nil then
+						new_actions[name][parser.r_type] = { ttl=parser.r_ttl, rdata=rdata }
+					else -- mutiple RRs: no reordering or deduplication
+						if type(act.rdata) ~= 'table' then
+							act.rdata = { act.rdata }
+						end
+						table.insert(act.rdata, rdata)
+						if parser.r_ttl ~= act.ttl then -- be conservative
+							log_warn(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d warning: different TTLs in a set (minimum taken)',
+								path, tonumber(parser.line_counter))
+							act.ttl = math.min(act.ttl, parser.r_ttl)
+						end
+					end
+				else
+					assert(is_bad == false and prefix_labels == 0)
 				end
 			end
 		end
@@ -500,17 +512,15 @@ function policy.rpz(action, path, watch)
 					-- Watcher will also fire for changes to the directory itself
 					if name == file then
 						-- If the file changes then reparse and replace the existing ruleset
-						if verbose() then
-							log('[poli] RPZ reloading: ' .. name)
-						end
+						log_info(ffi.C.LOG_GRP_POLICY, 'RPZ reloading: ' .. name)
 						rules = rpz_parse(action, path)
 					end
 				end
 			end)
 		elseif watch then -- explicitly requested and failed
 			error('[poli] lua-cqueues required to watch and reload RPZ file')
-		elseif verbose() then
-			log('[poli] lua-cqueues required to watch and reload RPZ file, continuing without watching')
+		else
+			log_info(ffi.C.LOG_GRP_POLICY, 'lua-cqueues required to watch and reload RPZ file, continuing without watching')
 		end
 	end
 
@@ -598,7 +608,8 @@ local function answer_clear(req)
 	req.add_selected.len = 0
 
 	-- Let's be defensive and clear the answer, too.
-	local pkt = req.answer
+	local pkt = req:ensure_answer()
+	if pkt == nil then return nil end
 	pkt:clear_payload()
 	return pkt
 end
@@ -611,6 +622,7 @@ function policy.DENY_MSG(msg)
 	return function (_, req)
 		-- Write authority information
 		local answer = answer_clear(req)
+		if answer == nil then return nil end
 		ffi.C.kr_pkt_make_auth_header(answer)
 		answer:rcode(kres.rcode.NXDOMAIN)
 		answer:begin(kres.section.AUTHORITY)
@@ -632,24 +644,29 @@ end
 local debug_logline_cb = ffi.cast('trace_log_f', function (_, msg)
 	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
 	-- msg typically ends with newline
-	io.write(ffi.string(msg))
+	log_debug(ffi.C.LOG_GRP_POLICY, "%s", msg)
 end)
 ffi.gc(debug_logline_cb, free_cb)
 
 local debug_logfinish_cb = ffi.cast('trace_callback_f', function (req)
 	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
-	ffi.C.kr_log_req(req, 0, 0, 'dbg',
+	log_req(req, 0, 0, ffi.C.LOG_GRP_POLICY,
 		'following rrsets were marked as interesting:\n' ..
 		req:selected_tostring())
-	ffi.C.kr_log_req(req, 0, 0, 'dbg',
-		'answer packet:\n' ..
-		tostring(req.answer))
+	if req.answer ~= nil then
+		log_req(req, 0, 0, ffi.C.LOG_GRP_POLICY,
+			'answer packet:\n' .. tostring(req.answer))
+	else
+		log_req(req, 0, 0, ffi.C.LOG_GRP_POLICY,
+			'answer packet DROPPED\n')
+	end
 end)
 ffi.gc(debug_logfinish_cb, free_cb)
 
 -- log request packet
 function policy.REQTRACE(_, req)
-	ffi.C.kr_log_req(req, 0, 0, 'dbg', 'request packet:\n%s',
+	log_req(req, 0, 0, ffi.C.LOG_GRP_POLICY,
+		'request packet:\n%s',
 		tostring(req.qsource.packet))
 end
 
@@ -668,7 +685,7 @@ local debug_stashlog_cb = ffi.cast('trace_log_f', function (req, msg)
 end)
 ffi.gc(debug_stashlog_cb, free_cb)
 
--- buffer verbose logs and print then only if test() returns a truthy value
+-- buffer debug logs and print then only if test() returns a truthy value
 function policy.DEBUG_IF(test)
 	local debug_finish_cb = ffi.cast('trace_callback_f', function (cbreq)
 		jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
@@ -698,24 +715,27 @@ policy.DEBUG_CACHE_MISS = policy.DEBUG_IF(
 policy.DENY = policy.DENY_MSG() -- compatibility with < 2.0
 
 function policy.DROP(_, req)
-	answer_clear(req)
+	local answer = answer_clear(req)
+	if answer == nil then return nil end
 	return kres.FAIL
 end
 
 function policy.REFUSE(_, req)
 	local answer = answer_clear(req)
+	if answer == nil then return nil end
 	answer:rcode(kres.rcode.REFUSED)
 	answer:ad(false)
 	return kres.DONE
 end
 
 function policy.TC(state, req)
-	-- Skip non-UDP queries
-	if req.answer.max_size == 65535 then
+	-- Avoid non-UDP queries
+	if req.qsource.addr == nil or req.qsource.flags.tcp then
 		return state
 	end
 
 	local answer = answer_clear(req)
+	if answer == nil then return nil end
 	answer:tc(1)
 	answer:ad(false)
 	return kres.DONE
